from __future__ import annotations

import collections
import copy
import itertools
import warnings
from typing import Collection, Iterable


class RegexPattern:
    """接受正则表达式语法的一个子集（不支持 `.` ， `[^a]` 等任意能匹配字符的语法），
    展开为其所代表的字符串集合

    Examples
    --------
    'aaa|bbb|ccc' 展开为 ['aaa', 'bbb', 'ccc']

    'test[1234].(com|cn)/[012]{0,3}'

    Notes
    -----
    由于不同通配符展开结果会以笛卡尔积形式组合，警惕组合爆炸。

    [...]中的^会自动转义为该字符，不代表取反意义。

    `\\\\`依然为转义字符。

    字符集不包含大写字母表示的补集，例如 `\\S` 等。

    References
    ----------
    从字符串集合生成正则表达式
    http://regex.inginf.units.it/
    """
    EMPTY = ''

    def __init__(self, pattern: str, *, limit: int | bool | None = None, wide_spaces=False):
        self.pattern = pattern
        self.index = 0
        self.n_chars = len(self.pattern)

        if limit is True:
            limit = 100
        self.limit = limit

        self.char_classes = assemble_char_classes(wide_spaces=wide_spaces)

    def __iter__(self):
        if self.limit:
            yield from itertools.islice(self.parse(), self.limit)
        else:
            yield from self.parse()

    def fetch(self, escape=True):
        c = self.pattern[self.index]
        self.index += 1

        if c == '\\' and escape:
            c = self.escape_next()

        return c

    def escape_char_classes(self, c):
        return self.char_classes.get(c, None)

    def escape_next(self):
        c = self.fetch(escape=False)
        ret = self.escape_char_classes(c)
        if ret is None:
            ret = c
        return ret

    def parse(self, level=0):
        options = OrOp()
        components = JoinOp()

        while self.index < self.n_chars:
            c = self.fetch(escape=False)

            if c == '\\':
                components.append_char_or_component(self.escape_next())
            elif c == '(':
                components.seal_chars()
                components.append(self.parse(level=level + 1))
            elif c == '[':
                components.seal_chars()
                components.append(self.parse_charset())
            elif c in ('{', '?', '+', '*'):
                repeated = components.pop_last_char_and_seal_rest()
                if repeated is None:
                    if len(components) == 0:
                        raise SyntaxError('Repeat operator on empty operand.')
                    repeated = components.pop()
                components.append(self.parse_repetition(component=repeated, op=c))
            elif c == '|':
                components.seal_chars()
                options.append(components.unwrap())
                components = JoinOp()
            elif c == ')':
                if level <= 0:
                    raise SyntaxError('Brackets not enclosed.')
                break
            else:
                components.append_char(c)

        components.seal_chars()
        options.append(components.unwrap())
        return options.unwrap()

    def parse_charset(self):
        ret = CharsetOp()
        last = None

        while self.index < self.n_chars:
            c = self.fetch()

            if c == '-':
                c = self.fetch()

                if not isinstance(last, str) or not isinstance(c, str):
                    # 如果不是字符，那就是 `[` 或 `CharsetOp` ，是字符类转义
                    # 不应被 `-` 视为范围起点
                    start = last
                    if isinstance(start, CharsetOp):
                        start = start.clazz
                    elif start is None:
                        start = '['

                    end = c
                    if isinstance(end, CharsetOp):
                        end = end.clazz

                    raise SyntaxError(f'Unexpected `-` encountered between `{start}` and `{end}`.')

                if ord(last) > ord(c):
                    raise SyntaxError(f'Bad character range `{last}` and `{c}`.')
                ret.extend(chr(i) for i in range(ord(last), ord(c) + 1))
            elif c != ']':
                if c == char_class_empty:
                    # TODO: 明确这个搭配的语义
                    # \b 出现在范围语法中，似乎是视为空格
                    c = ' '
                ret.append_char_or_component(c)
            else:
                break

            last = c

        return ret

    def parse_repetition(self, component, op):
        if op == '{':
            values = []
            value = None
            while self.index < self.n_chars:
                c = self.fetch()

                if c == ',':
                    values.append(value)
                    value = None
                elif c == ' ':
                    continue
                elif c == '}':
                    values.append(value)
                    break
                else:
                    try:
                        v = int(c)

                        if value is None:
                            value = v
                        else:
                            value = value * 10 + v
                    except ValueError:
                        raise SyntaxError(f'Unexpected char `{c}` encountered.')

            if len(values) == 1:
                mmin, mmax = values[0], values[0]
            elif len(values) == 2:
                mmin, mmax = values
            else:
                raise SyntaxError(f'Invalid use of repeat.')

            if mmin is None:
                mmin = 0
        elif op == '?':
            mmin, mmax = 0, 1
        elif op == '+':
            mmin, mmax = 1, None
        elif op == '*':
            mmin, mmax = 0, None
        else:
            raise SyntaxError(f'Unknown repeat operator `{op}`.')

        if mmax is None and self.limit is None:
            warnings.warn(f'Unlimited quantifier `{op}` detected at {self.index},'
                          f' consider specifying `limit=100` to limit the output'
                          f' or `limit=False` to explicitly mute the warning message.')

        return RepeatOp(component, mintimes=mmin, maxtimes=mmax)


def expand_regex_pattern(pattern: str, limit: int | bool | None = None):
    pattern = RegexPattern(pattern, limit=limit)
    return pattern


class LiteralOp:
    def __init__(self, literal):
        self.literal = literal

    def __iter__(self):
        yield self.literal


class JoinOp:
    def __init__(self):
        self.components = []
        self.chars = ''

    def append(self, component):
        self.components.append(component)

    def pop(self):
        return self.components.pop()

    def append_char(self, c):
        self.chars += c

    def append_char_or_component(self, c):
        if isinstance(c, str):
            self.append_char(c)
        else:
            self.append(c)

    def seal_chars(self):
        if self.chars != '':
            self.append(LiteralOp(self.chars))
            self.chars = ''

    def pop_last_char_and_seal_rest(self):
        n_chars = len(self.chars)
        if n_chars > 1:
            self.append(LiteralOp(self.chars[:-1]))
            ret = self.chars[-1]
            self.chars = ''
        elif n_chars == 1:
            ret = self.chars
            self.chars = ''
        else:
            ret = None

        return ret

    def unwrap(self):
        if len(self.components) == 1:
            return self.components[0]
        else:
            return self

    def __len__(self):
        return len(self.components)

    def __iter__(self):
        for parts in lazy_product(*self.components):
            yield ''.join(parts)


class OrOp:
    def __init__(self):
        self.components = []

    def append(self, composition):
        self.components.append(composition)

    def unwrap(self):
        if len(self.components) == 1:
            return self.components[0]
        else:
            return self

    def __len__(self):
        return len(self.components)

    def __iter__(self):
        for c in self.components:
            yield from c


class CharsetOp:
    def __init__(self, init=None, clazz: str | None = None):
        if clazz is not None:
            if not clazz.startswith('\\'):
                clazz = '\\' + clazz

            if len(clazz) > 2:
                warnings.warn(f'`{CharsetOp.__name__}` got `{clazz}` as class identifier,'
                              f' taking it as {clazz[:2]}.')
                clazz = clazz[:2]
        self.clazz = clazz
        self.chars = collections.OrderedDict()
        if init is not None:
            self.extend(init)

    @property
    def class_identifier(self):
        return self.clazz[1:]

    def append(self, char):
        self.chars[char] = None

    def extend(self, string):
        for char in string:
            self.append(char)

    def append_char_or_component(self, c):
        if isinstance(c, str):
            self.extend(c)
        elif isinstance(c, CharsetOp):
            self.extend(c.chars)

    def __len__(self):
        return len(self.chars)

    def __iter__(self):
        yield from self.chars

    def __eq__(self, other):
        if not isinstance(other, CharsetOp):
            return False
        return self.clazz == other.clazz and self.chars == other.chars


class RepeatOp:
    def __init__(self, component, maxtimes: int | None, mintimes: int = 0):
        self.component = component
        self.mintimes = mintimes
        self.maxtimes = maxtimes

    def __iter__(self):
        base = [self.component]
        times = self.mintimes

        while self.maxtimes is None or times <= self.maxtimes:
            if times == 0:
                yield ''
            else:
                for parts in lazy_product(*(base * times)):
                    yield ''.join(parts)

            times += 1


def lazy_product(*iterables, all_collections=False):
    """真正的lazy版product

    `itertools.product` 要求在迭代前就将所有迭代器遍历一遍得到 `tuple` 序列，
    这样会导致无穷序列无法适用于此操作。

    而 `lazy_product` 则通过单独实现一个缓存，来实现扫描的同时创建序列。

    Parameters
    ----------
    iterables
        用于求解笛卡尔积的迭代器
    all_collections
        指示是否所有的序列都是有穷序列

    Returns
    -------
    cartesian_product
        多个迭代器的笛卡尔积结果
    """
    n_iterables = len(iterables)
    if n_iterables < 1:
        return

    iterable = iterables[0]
    if n_iterables == 1:
        yield from iterable
    elif n_iterables == 2 and isinstance(iterable, Collection):
        for elem0 in iterable:
            for elem1 in iterable:
                yield elem0, elem1
    else:
        others = iterables[1:]
        if all_collections or check_if_all_collections(others):
            for elem0 in iterable:
                for elems in lazy_product(*iterables[1:], all_collections=True):
                    yield elem0, *elems
        else:
            elements = []
            for elem0 in iterable:
                if len(elements) == 0:
                    for elems in lazy_product(*iterables[1:]):
                        elements.append(elems)
                        yield elem0, *elems
                else:
                    for elems in elements:
                        yield elem0, *elems


def check_if_all_collections(iterables: Iterable[Iterable]):
    return all(isinstance(iterable, Collection) for iterable in iterables)


spaces7 = ' \r\n\t\v\f'
spaces16 = ''.join([chr(c) for c in [
    160, 8232, 8233, 5760, 6158, 8192, 8193, 8194, 8195, 8196,
    8197, 8198, 8199, 8200, 8201, 8202, 8239, 8287, 12288, 65279
]])
numbers = ''.join([chr(i + ord('0')) for i in range(10)])
letters = ''.join([chr(i + ord('a')) for i in range(26)])

char_class_numbers = CharsetOp(init='0123456789', clazz=r'd')
char_class_spaces7 = CharsetOp(init=spaces7, clazz=r's')
char_class_spaces16 = CharsetOp(init=spaces7 + spaces16, clazz=r's')
char_class_identifiers = CharsetOp(init='_' + letters + letters.upper() + numbers, clazz=r'w')
char_class_empty = CharsetOp(init='', clazz=r'b')


def assemble_char_classes(wide_spaces=False):
    classes = [
        char_class_numbers,
        char_class_identifiers,
        char_class_empty
    ]
    if wide_spaces:
        classes.append(char_class_spaces16)
    else:
        classes.append(char_class_spaces7)

    return {
        cls.class_identifier: cls for cls in classes
    }



from functools import reduce


class ConnectedTriangleSurfaces:
    def __init__(self):
        """用于依据联通性（是否有相邻边）将多个三角面的顶点分组
        """
        self.point_groups: list[set[int]] = []

    def add(self, pts):
        vs = set(pts)
        candidates = []
        diffs = []

        # 考虑到面在各个局部体内的连续性，一般而言新的面会邻接最后一个点集或者属于新的点集合
        for group in reversed(self.point_groups):
            diff = vs.difference(group)
            if len(diff) <= 1:
                candidates.append(group)
                diffs.append(diff)

                if len(candidates) > 1:
                    break  # 因为针对三角面，所以一定只能同时邻接两个组

        n_adjacent = len(candidates)

        if n_adjacent == 0:
            target_group = set(pts)
            self.point_groups.append(target_group)
        elif n_adjacent == 1:
            assert len(diffs) == 1
            candidates[0].update(diffs.pop())
        else:
            for g in candidates[1:]:
                self.point_groups.remove(g)

            reduce(lambda x, y: x.update(y), candidates)

    def get_groups(self):
        return self.point_groups



import math
import warnings
from typing import Union

import numpy as np
import taichi as ti
from taichi.lang.misc import is_extension_supported
from taichi.lang.util import to_taichi_type

from metalpy.scab.utils.misc import Field
from metalpy.utils.numeric import limit_significand
from metalpy.utils.taichi import ti_kernel, ti_field, ti_FieldsBuilder, ti_func, copy_from, ti_size_t, ti_size_dtype, \
    ti_cfg, ti_pyfunc
from metalpy.utils.ti_solvers import matrix_free
from .kernel import kernel_matrix_forward
from .solver import DemagnetizationSolver


class CompressedSolver(DemagnetizationSolver):
    Optimal = 'optimal'

    def __init__(
            self,
            receiver_locations: np.ndarray,
            xn: np.ndarray,
            yn: np.ndarray,
            zn: np.ndarray,
            base_cell_sizes: np.ndarray,
            source_field: Field,
            compressed_size: Union[int, float, None] = None,
            deterministic: Union[bool, str] = True,
            quantized: bool = False,
            symmetric: bool = False,
            progress: bool = False
    ):
        """该函数将核矩阵通过哈希表进行压缩，以计算效率为代价换取大幅降低内存需求量

        Parameters
        ----------
        receiver_locations
            观测点
        xn, yn, zn
            网格边界
        base_cell_sizes
            网格最小单元大小
        compressed_size
            压缩表尺寸，
            如果为整数，则用于指定尺寸大小，
            如果为浮点数则设置为核矩阵大小的对应比例，
            默认None则为1 / 32
        deterministic
            采用确定性模式并行哈希，默认为True，牺牲一定计算效率和空间确保核矩阵正确。
            指定为CompressedForward.Optimal时直接采用unique函数计算理想压缩结果
        quantized
            采用量化模式压缩核矩阵，默认为False。
            True则启用压缩，使用更小位宽来存储核矩阵，减少一定的空间需求，略微牺牲计算效率。
            （可能会报要求尺寸为2的幂的性能警告）
        symmetric
            指示是否启用对称模式（当网格为规则网格时可以启用），默认为False。
            True则启用对称模式，以对称矩阵形式存储核矩阵，减少一定的空间需求
        progress
            是否输出求解进度条，默认为False不输出

        Notes
        -----
        优势：

        - 可以大幅降低内存需求量，对规则网格效果尤其明显，规则网格上约可以将内存需求降低到千分之一

        缺陷：

        - 网格规模仍然受到taichi的int32索引限制
        - 对计算效率有较大影响

        `deterministic` 指定是否采用确定性模式并行哈希，可选的取值为：

        - `True` （默认值），确保核矩阵正确，但会牺牲一定计算效率和空间
        - `False` 则会抛弃确定性约束，牺牲一定精度获取更优的时空间效率
        - `CompressedForward.Optimal` 时直接采用unique函数计算理想压缩结果

        `quantized` 指定是否对索引进行量化压缩，将一个整形拆分为多个值从而实现压缩，可选的取值为：

        - `False` 不启用压缩（默认值），采用默认索引类型存储（ti_size_t）
        - `True` 启用压缩，使用更小位宽来存储索引，从而在相同的空间中存储更多索引值，减少一定的空间需求

            在当前64位位宽下，只支持最多压缩到21位，索引约200万项，
            更宽则一个int64中只能存储2个索引，和32位相同，此时自动转为非压缩模式。

        `symmetric` 指定是否启用对称模式（当网格为规则网格时可以启用）存储索引矩阵，
        对称模式下，所有对角线元素值相同，存储在 indices_mat[0] ，上三角部分逐行存储在 indices_mat[1:] 。

        注意事项：

        - 网格尺寸建议采用 `IEEE754` 可以精确表示的值，防止因计算误差干扰压缩效果
        - 例如使用 0.375 或 0.4375 替代 0.4
        - 可以使用 `metalpy.utils.numeric.limit_significand` 来搜索合适的网格尺寸
        """
        super().__init__(receiver_locations, xn, yn, zn, base_cell_sizes, source_field)

        # 检查二进制小数的稳定性，例如 `0.4` 无法使用二进制小数精确表示，因此在其上的加减法会引入误差
        # 而CompressedSolver依赖数值的绝对相等来实现去重，对这种误差较为敏感，需要进行检查
        check_binary_floats_stability(
            receiver_locations - receiver_locations[0],
            threshold=0.95,
            axes_names='XYZ'
        )

        self.deterministic = deterministic
        self.progress = progress

        nC = xn.shape[0]
        nObs = receiver_locations.shape[0]

        default_table_size = get_default_table_size(nC)
        if compressed_size is None:
            compressed_size = default_table_size

        if np.issubdtype(type(compressed_size), np.floating):
            compressed_size = int(default_table_size * compressed_size)

        self.compressed_size = compressed_size
        self.used, self.overflow = 0, False

        self.builder = builder = ti_FieldsBuilder()

        self.Tmat321 = [
            [ti_field(self.kernel_type) for _ in range(3)],  # Txx, Txy, Txz
            [ti_field(self.kernel_type) for _ in range(2)],  # ---- Tyy, Tyz
            [ti_field(self.kernel_type) for _ in range(1)],  # --------- Tzz
        ]
        self.Tmat6 = [t for ts in self.Tmat321 for t in ts]

        builder.dense(ti.i, compressed_size).place(*self.Tmat6)

        self.symmetric = symmetric
        if not symmetric:
            axes = ti.ij
            indices_size = [nObs, nC]
        else:
            axes = ti.i
            indices_size = [symmetric_mat_size(nObs, nC)]

        self.quantized = quantized
        indices_mat = None
        while quantized:
            arch = ti_cfg().arch
            if not is_extension_supported(arch, ti.extension.quant):
                warnings.warn(f'`{arch.name}` does not support quantized types.'
                              f' Ignoring `quantized` and using {ti_size_dtype.__name__} instead.')
                break

            # TODO: 万一将来位宽变成128了？
            primitive_bits = np.iinfo(np.int64).bits
            quantized_bits = math.ceil(np.log2(compressed_size))
            n_packed = primitive_bits // quantized_bits
            if n_packed <= 2:
                warnings.warn(f'A single `uint{primitive_bits}` can pack only'
                              f' {n_packed} quantized index type `uint{quantized_bits}`.'
                              f' Ignoring `quantized` and using `{ti_size_dtype.__name__}` instead.')
                break

            indices_type = ti.types.quant.int(
                bits=quantized_bits,
                signed=False,
                compute=ti_size_t
            )

            indices_mat = ti_field(indices_type)
            indices_size[-1] = math.ceil(indices_size[-1] / n_packed)

            (
                builder
                .dense(axes, indices_size)
                .quant_array(axes[-1:], n_packed, max_num_bits=primitive_bits)
                .place(indices_mat)
            )

            break

        if indices_mat is None:
            indices_mat = ti_field(ti_size_t)
            builder.dense(axes, indices_size).place(indices_mat)

        self.indices_mat = indices_mat

        builder.finalize()

    def build_kernel(self, model):
        self.used, self.overflow = compress_kernel(
            self.Tmat6,
            self.receiver_locations,
            self.xn, self.yn, self.zn,
            self.indices_mat,
            deterministic=self.deterministic,
            symmetric=self.symmetric
        )

        kernel_matrix_forward(
            self.receiver_locations,
            self.xn, self.yn, self.zn,
            self.base_cell_sizes, model,
            *self.Tmat6, np.empty(0),
            write_to_mat=False, compressed=True
        )

    def solve(self, magnetization):
        solver = not self.symmetric and solve_Tx_b_compressed or solve_Tx_b_compressed_symmetric
        return solver(self.Tmat321, self.indices_mat, magnetization, progress=self.progress)


def search_prime(value):
    """搜索合适的质数，默认从good primes中取，备选选项为更大的一个质数

    Parameters
    ----------
    value
        需要的质数的上界

    Returns
    -------
    prime
        小于value的某个合适的质数

    References
    ----------
    寻找哈希表的"好质数"： https://planetmath.org/goodhashtableprimes
    寻找下一个质数： https://www.numberempire.com/primenumbers.php
    """
    primes = {
        1 << 0: [1, 1],  # 别问
        1 << 1: [1, 2],
        1 << 2: [2, 3],
        1 << 3: [5, 7],
        1 << 4: [11, 13],
        1 << 5: [23, 29],
        1 << 6: [53, 59],
        1 << 7: [97, 101],
        1 << 8: [193, 197],
        1 << 9: [389, 397],
        1 << 10: [769, 773],
        1 << 11: [1543, 1549],
        1 << 12: [3079, 3083],
        1 << 13: [6151, 6163],
        1 << 14: [12289, 12301],
        1 << 15: [24593, 24611],
        1 << 16: [49157, 49169],
        1 << 17: [98317, 98321],
        1 << 18: [196613, 196643],
        1 << 19: [393241, 393247],
        1 << 20: [786433, 786449],
        1 << 21: [1572869, 1572871],
        1 << 22: [3145739, 3145741],
        1 << 23: [6291469, 6291487],
        1 << 24: [12582917, 12582919],
        1 << 25: [25165843, 25165853],
        1 << 26: [50331653, 50331683],
        1 << 27: [100663319, 100663327],
        1 << 28: [201326611, 201326621],
        1 << 29: [402653189, 402653201],
        1 << 30: [805306457, 805306459],
        1 << 31: [1610612741, 1610612747],
    }

    key = 1 << int(np.log2(value))
    candidates = primes[key]
    if candidates[0] == value:
        return candidates[1]
    else:
        return candidates[0]


def compress_kernel(
        Tmat6,
        receiver_locations: np.ndarray,
        xn: np.ndarray,
        yn: np.ndarray,
        zn: np.ndarray,
        indices_mat,
        deterministic,
        symmetric
):
    if deterministic not in [True, False]:
        assert deterministic == CompressedSolver.Optimal

    table_size = Tmat6[0].shape[0]
    default_table_size = get_default_table_size(xn.shape[0])

    if deterministic == CompressedSolver.Optimal:
        used, overflow = compress_kernel_optimal(
            Tmat6, receiver_locations,
            xn, yn, zn,
            indices_mat,
            symmetric
        )
    else:
        used, overflow = compress_kernel_by_hash(
            Tmat6, receiver_locations,
            xn, yn, zn,
            indices_mat,
            deterministic,
            symmetric
        )

    if overflow:
        raise RuntimeError(f'Compression hash table overflowed.'
                           f' Consider using larger `compressed_size` like'
                           f' `{int(table_size * 1.5)}` or'
                           f' `{table_size / default_table_size * 1.5:.2f}`'
                           f' (currently {table_size}).')

    memory_efficiency = used / table_size
    if memory_efficiency < 0.95:
        # 一个特点是，尤其是对于规则六边形网格
        # 不同网格关系数的和有效网格数近似线性关系
        warnings.warn(f'Low memory efficiency detected ({memory_efficiency:.2%}).'
                      f' Consider setting `compressed_size` to'
                      f' `{int(used / 0.95)}`'
                      f' or'
                      f' `{used / 0.95 / default_table_size:.4f}`'
                      f'.')

    return used, overflow > 0


def compress_kernel_optimal(
        mats,
        receiver_locations: np.ndarray,
        xn: np.ndarray,
        yn: np.ndarray,
        zn: np.ndarray,
        indices_mat,
        symmetric: bool
):
    n_param = len(mats)
    table_size = mats[0].shape[0]

    table = np.c_[xn, yn, zn][:, None, :] - np.repeat(receiver_locations, 2, axis=1)[None, :, :]
    compressed, inverse = np.unique(table.reshape(-1, n_param), axis=0, return_inverse=True)

    used = compressed.shape[0]
    overflow = used > table_size

    if not overflow:
        for i in range(n_param):
            copy_col_vector_from_matrix(mats[i], compressed, i)

        inverse = inverse.reshape(table.shape[:2]).astype(ti_size_dtype)
        if not symmetric:
            indices_mat.from_numpy(inverse)
        else:
            indices_mat.from_numpy(inverse[np.triu_indices(xn.shape[0])])

    return used, overflow > 0


def compress_kernel_by_hash(
        mats,
        receiver_locations: np.ndarray,
        xn: np.ndarray,
        yn: np.ndarray,
        zn: np.ndarray,
        indices_mat,
        deterministic: bool,
        symmetric: bool
):
    table_size = mats[0].shape[0]

    # TODO: 等taichi更新获取变量类型的泛型操作后可以简化
    val_type = np.result_type(receiver_locations, xn)
    assert np.issubdtype(val_type, np.floating)
    val_type = getattr(np, str(val_type).replace('float', 'uint'))
    total_bits = np.iinfo(val_type).bits

    overflow = ti_field(ti.i8, 1)
    prime = search_prime(table_size)

    if deterministic:
        # 使用原子操作来防止哈希表的单个条目被多个线程操作，确保核矩阵参数正确
        # 但会导致额外的计算和存储开销，实际压缩后尺寸会大于等于理想压缩结果
        #   GPU下原子操作可能会导致相同参数被存在多个条目下，一般结果大于理想压缩结果
        #   CPU下原子操作没有问题，一般结果等于理想压缩结果
        invalid_value_for_dx1 = 2 * (np.max(receiver_locations[:, 0]) - np.min(xn[:, 0]))
    else:
        # 小于0时hash_unique会取消采用原子操作，允许并行访问操作哈希表
        # 导致核矩阵部分参数产生偏差，实际压缩后尺寸会小于理想压缩结果
        invalid_value_for_dx1 = -1

    hash_unique(
        receiver_locations,
        xn, yn, zn,
        *mats,
        indices_mat,
        to_taichi_type(val_type),
        total_bits,
        overflow,
        prime,
        invalid_value_for_dx1,
        symmetric
    )

    overflow = overflow.to_numpy()[0]
    used = np.count_nonzero(mats[0].to_numpy() < invalid_value_for_dx1)

    return used, overflow > 0


@ti_kernel
def hash_unique(
        receiver_locations: ti.types.ndarray(),
        xn: ti.types.ndarray(),
        yn: ti.types.ndarray(),
        zn: ti.types.ndarray(),
        Txx: ti.template(),
        Txy: ti.template(),
        Txz: ti.template(),
        Tyy: ti.template(),
        Tyz: ti.template(),
        Tzz: ti.template(),
        indices_mat: ti.template(),
        hash_type: ti.template(),
        total_bits: ti.template(),
        overflow: ti.template(),
        prime: ti.template(),
        invalid_value_for_dx1: ti.template(),
        symmetric: ti.template()
):
    for I in ti.grouped(Txx):
        Txx[I] = ti.math.inf

    nC = xn.shape[0]
    nObs = receiver_locations.shape[0]
    table_size = Txx.shape[0]
    overflow[0] = ti.i8(0)
    qb = total_bits // 4  # quarter bits

    total = nObs * nC
    if ti.static(symmetric):
        total = symmetric_mat_size(nObs, nC)

    for i in ti.ndrange(total):
        if overflow[0] > 0:
            continue

        iobs, icell = i // nC, i % nC
        if ti.static(symmetric):
            iobs, icell = index_triu2mat(i, nC)

        dz2 = zn[icell, 1] - receiver_locations[iobs, 2]
        dz1 = zn[icell, 0] - receiver_locations[iobs, 2]
        dy2 = yn[icell, 1] - receiver_locations[iobs, 1]
        dy1 = yn[icell, 0] - receiver_locations[iobs, 1]
        dx2 = xn[icell, 1] - receiver_locations[iobs, 0]
        dx1 = xn[icell, 0] - receiver_locations[iobs, 0]

        if ti.static(not symmetric):
            # 由于互换对称性，核矩阵会存在一定的对称性
            # 极端情况是规则网格，此时核矩阵为对称矩阵，启用对称模式即不再需要该trick

            # 因此通过确保网格关系的唯一性，可以最高将压缩率提高到原来的50%
            # 唯一性条件具体以dx条件为例：
            #   情况1. dx1和dx2同号，保证均为正
            #   情况2. dx1和dx2异号，保证dx1绝对值小于dx2 （注：该情况对于规则网格无意义）
            # 检查过程如下：
            # 1. 检查dx条件，不满足则执行互换
            # 2. 如果dx相反（dx1和dx2互为相反数，此时dx1和dx2互换前后数值不变），则检查dy条件，不满足则执行互换
            # 3. 如果dx和dy均相反，则检查dz条件，不满足则执行互换
            # 提升比需要在CPU模式下验证（对于GPU，由于冲突的存在，无法达到理想提升比）
            e1, e2 = dx1 + dx2 == 0, dy1 + dy2 == 0
            swap_x = dx1 < 0 and -dx1 > dx2  # 检查dx条件
            swap_y = e1 and dy1 < 0 and -dy1 > dy2  # dx相反，检查dy条件
            swap_z = e1 and e2 and dz1 < 0 and -dz1 > dz2  # dx和dy相反，检查dz条件

            if swap_x or swap_y or swap_z:
                dx1, dx2 = -dx2, -dx1
                dy1, dy2 = -dy2, -dy1
                dz1, dz2 = -dz2, -dz1

        p0 = ti.bit_cast(dx1, hash_type)
        p1 = ti.bit_cast(dx2, hash_type)
        p2 = ti.bit_cast(dy1, hash_type)
        p3 = ti.bit_cast(dy2, hash_type)
        p4 = ti.bit_cast(dz1, hash_type)
        p5 = ti.bit_cast(dz2, hash_type)

        # 哈希函数设计原则：
        # 由于p[i]是直接由d[xyz][12]从二进制角度转换过来
        # 因此其中指数部分（高位）应在大部分情况下相同，导致哈希空间稀疏，
        # 所以需要采用循环位移来使其充分混淆
        a0: ti_size_t = (
            cshift(p1, qb * 0)
            ^ cshift(p4, qb * 1)
            ^ cshift(p2, qb * 2)
            ^ cshift(p5, qb * 3)
            ^ cshift(p3, qb * 0)
            ^ cshift(p0, qb * 1)
        ) % table_size

        # 预设了网格数大于并行线程数，因此同时查询的键几乎不会重复
        # 访问冲突的概率只取决于哈希函数的碰撞率
        a = a0
        while True:
            if ti.static(invalid_value_for_dx1 > 0):
                # 严格模式，严格防止数据竞争
                # 1. 有可能导致相同参数占用多个哈希条目，导致额外内存开销
                # 2. 依赖原子操作，产生额外计算开销
                # TODO: 替换为ti.atomic_cas
                #   https://github.com/taichi-dev/taichi/issues/1805
                if ti.math.isinf(ti.atomic_min(Txx[a], invalid_value_for_dx1)):
                    break
            else:
                # 非严格模式，有可能因并发修改哈希表条目，导致少量网格参数覆盖
                if ti.math.isinf(Txx[a]):
                    break

            if (
                Txx[a] == dx1
                and Txy[a] == dx2
                and Txz[a] == dy1
                and Tyy[a] == dy2
                and Tyz[a] == dz1
                and Tzz[a] == dz2
            ):
                break

            # 采用与nC/nObs互质的距离来做跳跃rehash：
            # 降低二次冲突概率，同时保证能遍历全表
            a += prime
            if a >= table_size:
                a -= table_size

            if a == a0:
                # 无法解决冲突，认为哈希表已经溢出
                overflow[0] = ti.i8(1)
                break

        # 容量超了的话，其实已经不需要管下面数据错误的问题了
        # if overflow[0] > 0:
        #     # print('寄！')
        #     continue

        if ti.static(not symmetric):
            indices_mat[iobs, icell] = a
        else:
            indices_mat[i] = a

        Txx[a] = dx1
        Txy[a] = dx2
        Txz[a] = dy1
        Tyy[a] = dy2
        Tyz[a] = dz1
        Tzz[a] = dz2


@ti_func
def cshift(key, b):
    return (key << b) | (key >> b)


def solve_Tx_b_compressed(Tmat321, indices_mat, m, progress: bool = False):
    with ti_FieldsBuilder() as builder:
        x = builder.place_dense_like(m[:, None])
        b = builder.place_dense_like(m[:, None])

        builder.finalize()

        copy_from(b, m[:, None])

        dtype = x.dtype

        @ti_func
        def extract(r, c):
            ind = indices_mat[r, c]
            return (
                Tmat321[0][0][ind], Tmat321[0][1][ind], Tmat321[0][2][ind],
                Tmat321[1][0][ind], Tmat321[1][1][ind],
                Tmat321[2][0][ind]
            )

        @ti_kernel
        def linear(x: ti.template(), Ax: ti.template()):
            n_obs = Ax.shape[0] // 3
            n_cells = x.shape[0] // 3

            for r in range(n_obs):
                row = r * 3

                bx: dtype = 0
                by: dtype = 0
                bz: dtype = 0

                for c in range(0, n_cells):
                    col = c * 3

                    txx, txy, txz, tyy, tyz, tzz = extract(r, c)
                    mx, my, mz = x[col + 0, 0], x[col + 1, 0], x[col + 2, 0]

                    bx += txx * mx + txy * my + txz * mz
                    by += txy * mx + tyy * my + tyz * mz
                    bz += txz * mx + tyz * my + tzz * mz

                Ax[row + 0, 0] = bx
                Ax[row + 1, 0] = by
                Ax[row + 2, 0] = bz

        matrix_free.cg(ti.linalg.LinearOperator(linear), b, x, progress=progress)

        return x.to_numpy()


def solve_Tx_b_compressed_symmetric(Tmat321, indices_mat, m, progress: bool = False):
    with ti_FieldsBuilder() as builder:
        x = builder.place_dense_like(m[:, None])
        b = builder.place_dense_like(m[:, None])

        builder.finalize()

        copy_from(b, m[:, None])

        dtype = x.dtype

        n_cells = x.shape[0] // 3

        @ti_func
        def extract_by_index(i):
            ind = indices_mat[i]
            return (
                Tmat321[0][0][ind], Tmat321[0][1][ind], Tmat321[0][2][ind],
                Tmat321[1][0][ind], Tmat321[1][1][ind],
                Tmat321[2][0][ind]
            )

        @ti_func
        def extract(r, c):
            i = index_mat2triu(r, c, n_cells)
            return extract_by_index(i)

        @ti_func
        def extract_diag():
            return extract_by_index(0)

        @ti_kernel
        def linear(x: ti.template(), Ax: ti.template()):
            n_obs = Ax.shape[0] // 3
            n_cells = x.shape[0] // 3

            # 对角线和上三角部分
            txx_, txy_, txz_, tyy_, tyz_, tzz_ = extract_diag()
            for r in range(n_obs):
                row = r * 3

                mx_, my_, mz_ = x[row + 0, 0], x[row + 1, 0], x[row + 2, 0]

                bx_u: dtype = txx_ * mx_ + txy_ * my_ + txz_ * mz_
                by_u: dtype = txy_ * mx_ + tyy_ * my_ + tyz_ * mz_
                bz_u: dtype = txz_ * mx_ + tyz_ * my_ + tzz_ * mz_

                for c in range(r + 1, n_cells):
                    col = c * 3

                    txx, txy, txz, tyy, tyz, tzz = extract(r, c)
                    mx, my, mz = x[col + 0, 0], x[col + 1, 0], x[col + 2, 0]

                    bx_u += txx * mx + txy * my + txz * mz
                    by_u += txy * mx + tyy * my + tyz * mz
                    bz_u += txz * mx + tyz * my + tzz * mz

                Ax[row + 0, 0] = bx_u
                Ax[row + 1, 0] = by_u
                Ax[row + 2, 0] = bz_u

            # 下三角部分
            for c in range(1, n_cells):
                col = c * 3

                bx_l: dtype = 0
                by_l: dtype = 0
                bz_l: dtype = 0

                for r in range(c):
                    row = r * 3

                    txx, txy, txz, tyy, tyz, tzz = extract(r, c)
                    mx, my, mz = x[row + 0, 0], x[row + 1, 0], x[row + 2, 0]

                    bx_l += txx * mx + txy * my + txz * mz
                    by_l += txy * mx + tyy * my + tyz * mz
                    bz_l += txz * mx + tyz * my + tzz * mz

                Ax[col + 0, 0] += bx_l
                Ax[col + 1, 0] += by_l
                Ax[col + 2, 0] += bz_l

        matrix_free.cg(ti.linalg.LinearOperator(linear), b, x, progress=progress)

        return x.to_numpy()


@ti_kernel
def copy_col_vector_from_matrix(
        dst: ti.template(), src: ti.types.ndarray(),
        col: ti.template()
):
    size = ti.min(dst.shape[0], src.shape[0])
    for i in range(size):
        dst[i] = src[i, col]


def check_binary_floats_stability(matrix, threshold=0.95, axes_names=None):
    """检查二进制小数的稳定性，例如 `0.4` 无法使用二进制小数精确表示，
    因此在其上的运算会引入误差，误差累积后会导致结果发生细微的偏差

    Parameters
    ----------
    matrix
        需要检查稳定性的矩阵
    threshold
        稳定性阈值（0 ~ 1，列元素中稳定的元素数大于该比例则认为合格）
    axes_names
        用于输出的坐标轴名
    """
    stable_mask = np.equal(matrix, limit_significand(matrix))
    axes_ratio = np.count_nonzero(stable_mask, axis=0) / stable_mask.shape[0]
    for axis, rate in enumerate(axes_ratio):
        if rate < threshold:
            if axes_names is not None:
                axis = axes_names[axis]
            warnings.warn(f'Low numeric stability detected on axis {axis}'
                          f' ({rate:.2%} < {threshold:.2%}),'
                          f' which may severely affect compression rate.'
                          f' Consider adjusting mesh cell sizes by'
                          f' `limit_significand(your_cell_sizes)`.'
                          f'\n'
                          f' See `metalpy.utils.numeric.limit_significand`'
                          f' for more details.')


def get_default_table_size(n_cells):
    return 100 * n_cells


@ti_pyfunc
def mul_and_div2(a, b):
    """计算 a * b // 2，通过调换除法顺序来防止溢出
    """
    p = a & 1  # a是否是2的倍数
    a >>= 1 - p
    b >>= p

    return a * b


@ti_pyfunc
def symmetric_mat_size(nobs, nc):
    """计算对称模式下的存储大小，包含一个对角线元素和上三角部分
    """
    # nobs * (nc - 1) // 2 + 1
    return mul_and_div2(nobs, nc - 1) + 1


@ti_func
def index_triu2mat(i, n):
    """通过索引号i计算上三角矩阵下的行列号

    i = 0 时代表对角线元素，返回 c = r = 0

    i > 0 时计算对应上三角部分的行列号 c 和 r
    """
    r: ti_size_t = 0
    c: ti_size_t = 0

    if i != 0:
        # ref: https://stackoverflow.com/a/69292749
        # 注意此处 i=0 代表所有对角线元素，因此对应到回答中 k=i-1
        # 但是由于计算量较大且存在精度损失，效率反而不如迭代计算
        # r = n - 2 - ti.math.floor(ti.math.sqrt(4.0 * n * (n - 1) - (8 * (i - 1)) - 7) / 2.0 - 0.5, dtype=ti_size_t)

        r = 0
        c = i
        while c >= n:
            r = r + 1 + (c - (n - 1)) // ((n - 1) - (r + 1))
            c = i - mul_and_div2(2 * n - (r + 1), r) + r
        while c <= r:
            c = c + n - (r + 1)
            r -= 1

    return r, c


@ti_pyfunc
def index_mat2triu(r, c, n):
    """通过上三角矩阵下的行列号计算索引号i，需要调用方保证 c > r
    """
    # n * r - r * (r + 1) // 2 + (c - (r + 1)) + 1
    # 可以变形为
    # (2 * n - (r + 1)) * r // 2 + (c - (r + 1)) + 1
    return mul_and_div2(2 * n - (r + 1), r) + (c - (r + 1)) + 1



from __future__ import annotations

import re
import warnings
from typing import Literal, Callable, Iterable, TypeVar, Sequence

_T = TypeVar('_T')


def replace_batch(string, mapping):
    keys = (re.escape(k) for k in mapping.keys())
    pattern = re.compile('(' + '|'.join(keys) + ')')
    result = pattern.sub(lambda x: mapping[x.group()], string)

    return result


def format_string_list(strs,
                       multiline: int | bool = 15,
                       limit: int | Literal[False] = 5,
                       line_indent=4,
                       quote: str | Callable | None = None):
    """将给定字符串数组格式化为列表形式

    Parameters
    ----------
    strs
        字符串列表
    multiline
        是否允许在字符串元素较长时格式化为多行列表，也可通过指定一个数值来限制换行的大小
    limit
        最大将输出的元素个数，False代表不限制输出元素个数
    line_indent
        触发格式化为多行列表时，行的缩进字符数
    quote
        对每个字符串两端添加字符，None代表不作处理。
        若为字符串，长度为1时两端添加相同字符，长度为2时分别加在两端。

    Returns
    -------
    formatted_str
        字符串数组格式化后的结果

    Notes
    -----
    multiline若为True，则采用默认长度限制15；
    若为False，则不会格式化为多行列表；
    若为int，则在元素长度超过该值时格式化为多行列表。

    转化为多行列表时会自动在字符串前添加换行符。

    Examples
    --------
    >>> print(format_string_list(['`Grav.`', '`Mag.`', '`EM`'], multiline=False, limit=1))
    <<< `Grav.` and ... (2 more)

    >>> print(format_string_list(['`Grav.`', '`Mag.`', '`EM`'], multiline=True, limit=1))
    <<<
    <<<     - `Grav.`
    <<<     - ... (2 more)

    >>> print(format_string_list(['`Grav.`', '`Mag.`', '`EM`'], multiline=False, limit=False))
    <<< `Grav.`, `Mag.` and `EM`

    >>> print(format_string_list(['`Grav.`', '`Mag.`', '`EM`'], multiline=True, limit=False))
    <<<
    <<<     - `Grav.`
    <<<     - `Mag.`
    <<<     - `EM`
    """
    if quote is not None:
        if isinstance(quote, str):
            if len(quote) == 1:
                lq, rq = quote, quote
            elif len(quote) == 2:
                lq, rq = quote
            else:
                warnings.warn(f'`format_string_list` with more than two `quote`s: {quote}')
                lq, rq = quote[:2]
            strs = [lq + s + rq for s in strs]
        elif callable(quote):
            strs = [quote(s) for s in strs]
        else:
            warnings.warn(f'`format_string_list` got unknown `quote`: {quote}.'
                          f' Expect `str` or `Callable`.'
                          f' Ignoring it.')
    else:
        if not isinstance(strs, list):
            strs = list(strs)

    if not isinstance(multiline, bool):
        line_limit = multiline
        if line_limit >= 0:
            max_length = max([len(s) for s in strs])
            multiline = max_length > line_limit
        else:
            multiline = True

    formatted = strs
    if limit:
        n_omitted = len(strs) - limit
        if n_omitted > 0:
            formatted = [*strs[:limit], f'... ({n_omitted} more)']

    if multiline:
        prefix = ' ' * line_indent + '- '
        return '\n' + '\n'.join([prefix + s for s in formatted])
    else:
        if len(formatted) > 1:
            return ', '.join(formatted[:-1]) + ' and ' + formatted[-1]
        else:
            return formatted[0]


def parse_axes_labels(
        labels: Iterable[str | int] | int,
        max_length: int | None = None,
        length: int | None = None
):
    return parse_labels(labels=labels, accepts='xyz', max_length=max_length, length=length)


def parse_labels(
        labels: Iterable[_T | int] | _T | int,
        accepts: Sequence[_T],
        max_length: int | None = None,
        length: int | None = None
):
    """解析一组字符串或数组为下标数组，例如将 'xzy' 解析为 [0, 2, 1]。

    Parameters
    ----------
    labels
        待解析的labels
    accepts
        指定接受的字符列表
    max_length
        指定接受的最大标签数
    length
        指定接受的标签数，多余或少于均拒绝

    Returns
    -------
    parsed_indices
        根绝给定的字符或整数列表解析的下标列表

    Examples
    --------
    >>> print(parse_axes_labels('xzy'))
    <<< [0, 2, 1]
    """
    max_idx = len(accepts)
    error_msg = f'`parse_axes_labels` accepts only' \
                f' {format_string_list(accepts, multiline=False)}' \
                f' or integer number < {max_idx}. Got `{{got}}`.'

    if length is not None:
        max_length = length

    if not isinstance(labels, Iterable):
        labels = (labels,)

    ret = []

    for c in labels:
        if isinstance(c, int):
            idx = c
        else:
            try:
                idx = accepts.index(c)
            except ValueError:
                idx = max_idx  # 找不到喵
        assert idx < max_idx, error_msg.format(got=c)
        ret.append(idx)

        if max_length is not None:
            assert len(ret) <= max_length, 'Too many labels.'

    if length is not None:
        assert len(ret) == length, f'Exactly {length} label(s) are required.'

    return ret



import numpy as np
from numpy.testing import assert_almost_equal

from metalpy.scab.modelling.transform import Rotation


def test_rotation():
    mesh = np.asarray([
        [0, 0, 1],
        [0, 0, 2],
        [0, 1, 0],
    ])
    rot = Rotation(30, 40, 50, degrees=True)
    transformed = rot.inverse_transform(rot.transform(mesh))
    assert_almost_equal(transformed, mesh)

    mesh = np.asarray([1, 0, 0])

    rot = Rotation(45, 90, -45, degrees=True, seq='yzx')
    assert_almost_equal(rot.transform(mesh), np.r_[0, 0, -1])
    assert_almost_equal(rot.inverse_transform(mesh), np.r_[0, -1, 0])



from __future__ import annotations

import warnings

import numpy as np
import psutil
import taichi as ti
from discretize.base import BaseTensorMesh
from discretize.utils import mkvc

from metalpy.scab.utils.misc import Field
from metalpy.utils.string import format_string_list
from metalpy.utils.taichi import ti_cfg, ti_test_snode_support, ti_size_max
from .solvers.compressed import CompressedSolver
from .solvers.integrated import IntegratedSolver
from .solvers.seperated import SeperatedSolver
from .solvers.solver import DemagnetizationSolver


class Demagnetization:
    Compressed = 'Compressed'
    Seperated = 'Seperated'
    Integrated = 'Integrated'
    methods = {
        Compressed,
        Seperated,
        Integrated,
    }

    def __init__(
            self,
            mesh: BaseTensorMesh,
            source_field: Field | None = None,
            active_ind=None,
            method=None,
            compressed_size: int | float | None = None,
            deterministic: bool | str = True,
            quantized: bool = False,
            symmetric: bool | None = None,
            progress=False
    ):
        """
        通过共轭梯度法求解计算退磁作用下的三轴等效磁化率

        Parameters
        ----------
        mesh
            模型网格
        source_field
            外部场源，也可以在dpred时指定，主要为保持和Simulation的一致性
        active_ind
            有效网格下标或掩码
        method
            采用的数值求解算法，参考 `Demagnetization.methods` 。默认为None，自动选择。
            可选选项：
                - Demagnetization.Compressed
                - Demagnetization.Seperated
                - Demagnetization.Integrated
        compressed_size
            压缩表尺寸，只针对`Demagnetization.Compressed`有效，用于指定压缩后的核矩阵哈希表最大大小，应尽可能小以满足内存要求。
            浮点数代表按完整核矩阵大小的比例设置，整数代表直接设置尺寸。
            如果大小不足或过大内部会给出提示信息
        deterministic
            采用确定性模式并行哈希，默认为True，牺牲一定计算效率和空间确保核矩阵正确。
            指定为CompressedForward.Optimal时直接采用unique函数计算理想压缩结果
        quantized
            采用量化模式压缩核矩阵，默认为False。
            True则启用压缩，使用更小位宽来存储核矩阵，减少一定的空间需求，略微牺牲计算效率。
            （可能会报要求尺寸为2的幂的性能警告）
        symmetric
            指示是否启用对称模式（当网格为规则网格时可以启用）。
            True则在检测到网格符合条件时启用对称模式，以对称矩阵形式存储核矩阵，减少一定的空间需求。
            False则禁用对称模式，无论是否符合条件。
            在CPU模式下会对性能造成较大影响，因此默认不启用。
            在GPU模式下性能影响较小，因此默认启用。
        progress
            是否输出求解进度条，默认为False不输出

        Notes
        -----
        compressed_size参考 `metalpy.scab.demag.solvers.compressed.CompressedSolver` 定义

        三个方法具体为：

        - `Integrated` 采用朴素算法，直接求解完整核矩阵
        - `Seperated` 将核矩阵拆分为9份来一定程度上绕过taichi对矩阵大小的限制
        - `Compressed` 通过哈希表将核矩阵压缩，以计算效率为代价换取大幅降低内存需求量

        网格规模较小时 `Integrated` 计算效率最高，
        随着网格规模增加 `Integrated` 和 `Seperated` 计算耗时持平，
        但随着网格规模继续增加， `Integrated` 和 `Seperated` 的内存需求量相对于网格数平方级膨胀，
        如果超出内存限制，则需要采用 `Compressed` 来时间换空间。
        """
        super().__init__()
        assert method is None or method in Demagnetization.methods, \
            f'`method` must by one of {format_string_list(Demagnetization.methods)}, got `{method}`.'

        self.mesh = mesh
        self.source_field = source_field
        self.method = method
        self.compressed_size = compressed_size
        self.deterministic = deterministic
        self.quantized = quantized
        self.progress = progress

        cell_centers = mesh.cell_centers
        if active_ind is not None:
            cell_centers = cell_centers[active_ind]

        self.receiver_locations = cell_centers

        # 计算网格在三个方向的边界位置
        h_gridded = mesh.h_gridded[active_ind]
        bsw = cell_centers - h_gridded / 2.0
        tne = cell_centers + h_gridded / 2.0

        xn1, xn2 = bsw[:, 0], tne[:, 0]
        yn1, yn2 = bsw[:, 1], tne[:, 1]
        zn1, zn2 = bsw[:, 2], tne[:, 2]

        self.Xn = np.c_[mkvc(xn1), mkvc(xn2)]
        self.Yn = np.c_[mkvc(yn1), mkvc(yn2)]
        self.Zn = np.c_[mkvc(zn1), mkvc(zn2)]

        base_cell_sizes = np.r_[
            self.mesh.h[0].min(),
            self.mesh.h[1].min(),
            self.mesh.h[2].min(),
        ]

        self.is_cpu = ti_cfg().arch == ti.cpu

        is_symmetric = all(np.all(h == h[0]) for h in mesh.h)
        if symmetric is None:
            symmetric = not self.is_cpu
        elif symmetric is True:
            if not is_symmetric:
                warnings.warn('`symmetric` mode is enabled, but the mesh is not symmetric.')
        self.symmetric = symmetric and is_symmetric

        # CPU后端可以直接判断内存是否足够
        # 但CUDA后端不能在没有额外依赖的情况下查询内存，并且内存不足时会抛出异常，无法通过再次ti.init还原
        # 因此留一个错误信息以供参考
        try:
            self.solver = dispatch_solver(
                self.receiver_locations, self.Xn, self.Yn, self.Zn, base_cell_sizes,
                source_field=self.source_field,
                is_cpu=self.is_cpu, method=self.method,
                compressed_size=self.compressed_size, deterministic=self.deterministic,
                quantized=self.quantized,
                symmetric=self.symmetric,
                progress=self.progress
            )
        except RuntimeError as e:
            if ti_cfg().arch == ti.cuda and 'cuStreamSynchronize' in e.args[0]:
                raise RuntimeError(f'Failed to build kernel matrix.'
                                   f' This may be resulted by insufficient memory'
                                   f' (please check log for details).'
                                   f'\nTry specifying `method={Demagnetization.__name__}.Compressed`'
                                   f' with proper `compressed_size` and'
                                   f' enabling `quantized` and `symmetric` when necessary.')
            else:
                raise e

    def dpred(self, model, source_field: Field | None = None):
        """计算退磁效应下的等效磁化率

        Parameters
        ----------
        model
            array-like(nC,)，磁化率模型
        source_field
            外部场源，覆盖求解器定义时给定的场源信息。
            如果为None，则采用给定的默认场源信息

        Returns
        -------
        ret
            array(nC, 3)，三轴等效磁化率矩阵
        """
        return self.solver.dpred(model, source_field=source_field).reshape(-1, 3)


def dispatch_solver(
        receiver_locations: np.ndarray,
        xn: np.ndarray,
        yn: np.ndarray,
        zn: np.ndarray,
        base_cell_sizes: np.ndarray,
        source_field,
        is_cpu=True,
        method=None,
        compressed_size: int | float | None = None,
        deterministic: bool | str = True,
        quantized: bool = True,
        symmetric: bool = False,
        progress=False
) -> DemagnetizationSolver:
    n_obs = receiver_locations.shape[0]
    n_cells = xn.shape[0]
    mat_size = n_obs * 3 * n_cells * 3
    kernel_size = mat_size * np.finfo(np.result_type(receiver_locations, xn)).bits / 8

    max_cells_allowed_0 = int((1 + np.sqrt(1 + 8 * ti_size_max)) / 2)  # CompressedSolver对称模式下支持的最大网格数
    max_cells_allowed_1 = int(np.sqrt(ti_size_max))  # CompressedSolver非对称模式和SeperatedSolver支持的最大网格数
    max_cells_allowed_2 = max_cells_allowed_1 // 3  # IntegratedSolver支持的最大网格数

    if is_cpu:
        available_memory = psutil.virtual_memory().free
    else:
        # TODO: 看taichi什么时候能出一个查询剩余显存/内存的接口
        available_memory = None

    common_kwargs = {
        'receiver_locations': receiver_locations,
        'xn': xn,
        'yn': yn,
        'zn': zn,
        'base_cell_sizes': base_cell_sizes,
        'source_field': source_field
    }
    kw_int = {'is_cpu': is_cpu, **common_kwargs}
    kw_sep = {'is_cpu': is_cpu, 'progress': progress, **common_kwargs}
    kw_com = {
        'compressed_size': compressed_size,
        'deterministic': deterministic,
        'quantized': quantized,
        'progress': progress,
        'symmetric': symmetric,
        **common_kwargs
    }

    if n_cells > max_cells_allowed_0:
        raise AssertionError(
            f'Mesh with {n_cells} cells (> {max_cells_allowed_0})'
            f' is not supported.'
        )
    elif n_cells > max_cells_allowed_1:
        assert method in {Demagnetization.Compressed, None}, (
            f'Mesh with {n_cells} cells (> {max_cells_allowed_1})'
            f' requires symmetric `{CompressedSolver.__name__}` (got `{solver_name(method)}`).'
        )
        assert symmetric, (
            f'Mesh with {n_cells} cells (> {max_cells_allowed_1})'
            f' requires `symmetric` mesh.'
            f' Try setting `symmetric=True`.'
        )
    elif n_cells > max_cells_allowed_2:
        assert method != Demagnetization.Integrated, (
            f'`{IntegratedSolver.__name__}` does not support'
            f' mesh with {n_cells} cells (> {max_cells_allowed_2}).'
        )

    if method is not None:
        candidate = dispatch_solver_by_name(
            method,
            kw_int=kw_int,
            kw_sep=kw_sep,
            kw_com=kw_com
        )
    elif available_memory is not None and kernel_size > available_memory * 0.8:
        candidate = CompressedSolver(**kw_com)
    elif n_cells > max_cells_allowed_1:
        candidate = CompressedSolver(**kw_com)
    elif n_cells > max_cells_allowed_2:
        candidate = SeperatedSolver(direct_to_host=ti_test_snode_support(), **kw_sep)
    else:
        print("Current GPU doesn't support SNode, fall back to legacy implementation.")
        candidate = IntegratedSolver(**kw_int)

    return candidate


def solver_name(method):
    return dispatch_solver_by_name(method).__name__


def dispatch_solver_by_name(method, kw_int=None, kw_sep=None, kw_com=None):
    """通过方法名获取求解器，当不指定求解器参数时，返回求解器类
    """
    if method == Demagnetization.Compressed:
        if kw_com is not None:
            candidate = CompressedSolver(**kw_com)
        else:
            candidate = CompressedSolver
    elif method == Demagnetization.Seperated:
        if kw_sep is not None:
            candidate = SeperatedSolver(direct_to_host=False, **kw_sep)
        else:
            candidate = SeperatedSolver
    elif method == Demagnetization.Integrated:
        if kw_int is not None:
            candidate = IntegratedSolver(**kw_int)
        else:
            candidate = IntegratedSolver
    else:
        warnings.warn('Unrecognized solver. Falling back to `Seperated` method.')
        if kw_sep is not None:
            candidate = SeperatedSolver(direct_to_host=False, **kw_sep)
        else:
            candidate = SeperatedSolver

    return candidate



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import overload


class WorkerContext(ABC):
    """实现消息通信上下文，通过序列化传递到Worker中，Worker调用fire给指定的event事件发送信息
    """

    @abstractmethod
    def fire(self, event, *args, **kwargs):
        pass

    @overload
    def bind(self, event) -> BoundWorkerContext: ...

    @overload
    def bind(self, event, idx: int | None) -> IndexedBoundWorkerContext: ...

    def bind(self, event, idx: int | None = None) -> BoundWorkerContext | IndexedBoundWorkerContext:
        if idx is None:
            return BoundWorkerContext(self, event)
        else:
            return IndexedBoundWorkerContext(self, event, idx=idx)


class BoundWorkerContext:
    def __init__(self, context: WorkerContext, event):
        """支持将WorkerContext绑定到给定事件上，从而在调用fire函数时不需要额外指定事件名

        Parameters
        ----------
        context
            事件消息通信上下文
        event
            绑定的事件名
        """
        self.context = context
        self.event = event

    def fire(self, *args, **kwargs):
        self.context.fire(self.event, *args, **kwargs)


class IndexedBoundWorkerContext(BoundWorkerContext):
    def __init__(self, context: WorkerContext, event, idx: int | None = None):
        """支持将WorkerContext绑定到给定事件上，从而在调用fire函数时不需要额外指定事件名

        Parameters
        ----------
        context
            事件消息通信上下文
        event
            绑定的事件名
        event
            绑定的事件回调函数的下标
        """
        super().__init__(context, event)
        self.idx = idx

    def __getstate__(self):
        """idx不需要一起传过去
        """
        return {
            'context': self.context,
            'event': self.event
        }



from metalpy.mexin.utils.type_map import assert_all_types_exists
from metalpy.scab.tied.tied import TaichiContext


def test_tied_impls_map():
    assert_all_types_exists(TaichiContext._implementations)



from typing import Mapping, Iterable

import numpy as np


def ensure_as_numpy_array_collection(array_collection):
    if isinstance(array_collection, dict):
        return {k: np.asarray(v) for k, v in array_collection.items()}
    elif isinstance(array_collection, list):
        return [np.asarray(v) for v in array_collection]
    elif isinstance(array_collection, tuple):
        return (np.asarray(v) for v in array_collection)
    elif isinstance(array_collection, Mapping):
        return {k: np.asarray(v) for k, v in array_collection.items()}
    elif isinstance(array_collection, Iterable):
        return [np.asarray(v) for v in array_collection]


def array_homogeneous_key(arr: np.ndarray):
    return arr.shape, arr.dtype


def is_homogeneous(inp: np.ndarray, out: np.ndarray):
    if inp.shape != out.shape:
        # 不允许数组尺寸发生改变
        return False
    if out.dtype != np.promote_types(inp.dtype, out.dtype):
        # 输出只允许类型抬升
        return False

    return True


class FixedShapeNDArray(np.ndarray):
    def __new__(cls, *args, **kwargs):
        return super().__new__(*args, **kwargs)

    def __array_finalize__(self, _, **__):
        pass

    def __array_ufunc__(self, ufunc, method, *inputs, out=None, **kwargs):
        typ = type(self)

        inputs = tuple(np.asarray(inp) if isinstance(inp, typ) else inp for inp in inputs)
        if out is not None:
            out = tuple(np.asarray(o) if isinstance(o, typ) else o for o in out)
        ret = super().__array_ufunc__(ufunc, method, *inputs, out=out, **kwargs)

        if ret is NotImplemented:
            return NotImplemented

        if is_homogeneous(self, ret):
            return ret.view(typ)
        else:
            return ret

    def reshape(
        self, *shape, **kwargs
    ):
        return self.view(np.ndarray).reshape(*shape, **kwargs)

    def __getitem__(self, item):
        return super().__getitem__(item).view(np.ndarray)



import threading
from typing import Iterable

from .mixed import Mixed
from .patch import Patch


class PatchContext:
    lock = threading.Lock()

    def __init__(self, *patches: Patch):
        """用于管理全局Patch的生命周期

        Parameters
        ----------
        patches
            待应用的patch

        Note
        ----
            进入时会将所有patch按照priority从小到大的顺序应用，退出时按逆序回滚

            由于作用于全局，因此需要加锁来防止多线程环境下造成patch混乱
        """
        self.patches = patches
        self.patches = sorted([*self.patches], key=lambda x: x.priority)
        self.mixers = []  # Mixed patches

    def __enter__(self):
        self.apply()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.rollback()

    def apply(self):
        PatchContext.lock.acquire()  # 防止多线程环境下造成patch混乱

        for patch in self.get_patches():
            patch.bind_context(self)
            patch.pre_apply()
            patch.apply()

        mixers = {}
        mixers_to_bind = []
        for patch in self.patches:
            for clazz, mixin_defs in patch.mixins.items():
                if clazz in mixers:
                    mixer = mixers[clazz]
                else:
                    mixer = Mixed(clazz)
                    mixers[clazz] = mixer
                    mixers_to_bind.append(mixer)

                for mixin_type, args, kwargs in mixin_defs:
                    mixer.mix(mixin_type, *args, **kwargs)

        for mixer in mixers_to_bind:
            mixer.bind_context(self)
            mixer.pre_apply()
            mixer.apply()

            self.mixers.append(mixer)

        for patch in self.get_patches_and_mixers():
            patch.commit()

    def rollback(self):
        for patch in self.get_patches_and_mixers(reverse=True):
            patch.rollback()
            patch.post_rollback()
            patch.unbind_context()

        self.mixers = []

        self.lock.release()

    def get_patches_and_mixers(self, reverse=False) -> Iterable[Patch]:
        """获取所有的patch（包括因mixin需要而对对应类加的Mixed patch）

        Returns
        -------
        ret
            所有的patch
        """
        if reverse:
            return reversed(self.mixers + self.patches)
        else:
            return self.mixers + self.patches

    def get_patches(self):
        """获取所有的patch（不包含因mixin需要而对对应类加的Mixed patch）

        Returns
        -------
        ret
            所有的patch
        """
        return self.patches


def patched(*patches):
    return PatchContext(*patches)



import sys

from SimPEG.simulation import BaseSimulation

from metalpy.mepa import Executor
from metalpy.mexin import Mixin
from metalpy.mexin.patch import Patch
from metalpy.mexin.injectors import reverted, replaces
from metalpy.mexin.lazy_class_delegate import LazyClassFactory
from .distributed_simulation import DistributedSimulation
from .utils import reget_class


class Distributed(Patch):
    def __init__(self, executor: Executor = None):
        super().__init__()
        self.executor = executor
        self.persisted_context = None

    def apply(self):
        self.add_mixin(BaseSimulation, Mixin)  # 保证BaseSimulation绑定mixin系统供DistributedSimulation使用

        if 'SimPEG.data' in sys.modules:
            from SimPEG import data
            self.hijack_cls_and_relative(data.Data, self._SimPEG_data_Data_wrapper)

        if 'SimPEG.potential_fields.magnetics.simulation' in sys.modules:
            import SimPEG.potential_fields.magnetics
            self.hijack_cls_and_relative(SimPEG.potential_fields.magnetics.simulation.Simulation3DIntegral,
                                         self._parallel_wrapper, executor=self.executor)
            self.hijack_cls_and_relative(SimPEG.potential_fields.magnetics.sources.UniformBackgroundField,
                                         self._lazy_wrapper)
            self.hijack_cls_and_relative(SimPEG.potential_fields.magnetics.sources.SourceField,
                                         self._lazy_wrapper)
            # Survey被人工重定向到了magnetics下，需要特殊处理
            self.hijack_cls_and_relative(SimPEG.potential_fields.magnetics.survey.Survey,
                                         self._lazy_wrapper, parent=False, child='survey')

    def hijack_cls_and_relative(self, cls, wrapper, parent=True, child=None, pass_cls=True, **extra_kwargs):
        """
        重定向类名字，使用wrapper函数替代
        :param cls: 目标类名
        :param wrapper: 替代的函数或者类

        :param parent: 是否在父级路径中重定向指定名字
        :param child: 是否在指定的子级路径中重定向指定名字
            这两个参数用于处理类的定义路径被修改的情形
            例如Survey定义在SimPEG.potential_fields.magnetics.survey，但被重定向到了SimPEG.potential_fields.magnetics
            需要同时重定向两个地方的定义
        :param pass_cls: 是否将cls作为第一个参数传递给wrapper
        :param extra_kwargs: 需要额外传给wrapper的参数
        """
        if pass_cls:
            extra_kwargs['cls'] = cls

        module_path = cls.__module__

        self.add_injector(replaces(cls, nest=sys.modules[module_path]),
                          lambda *args, **kwargs: wrapper(*args, **kwargs, **extra_kwargs))
        if parent:
            module_parent_path = '.'.join(module_path.split('.')[:-1])
            self.add_injector(replaces(cls, nest=sys.modules[module_parent_path]),
                              lambda *args, **kwargs: wrapper(*args, **kwargs, **extra_kwargs))
        if child is not None:
            module_child_path = '.'.join((module_path, child))
            self.add_injector(replaces(cls, nest=sys.modules[module_child_path]),
                              lambda *args, **kwargs: wrapper(*args, **kwargs, **extra_kwargs))

    def unbind_context(self):
        """重写unbind_context来延长PatchContext的生命周期
        """
        self.persisted_context = self.context
        super(Distributed, self).unbind_context()

    @property
    def priority(self):
        return 0

    def _parallel_wrapper(self, *args, cls, executor=None, **kwargs):
        return DistributedSimulation(self, cls, executor=executor, *args, **kwargs)

    @staticmethod
    def _lazy_wrapper(*args, cls, **kwargs):
        return LazyClassFactory(cls, *args, **kwargs)

    @staticmethod
    def _SimPEG_data_Data_wrapper(survey, dobs=None, cls=None, **kwargs):
        source_field = survey.find_param_by_type(LazyClassFactory, remove=False)

        # TODO: 可改进的HACK
        # 因为这些类被劫持了，得到的会是LazyClassFactory，
        # 因此需要通过获取Replacement对象临时还原才能正常构造(LazyClassFactory中的是原本的cls)
        repl_survey = reget_class(survey.cls)
        repl_sourcefield = reget_class(source_field.cls)
        with reverted(repl_survey, repl_sourcefield):  # 临时还原这两个被替换的类才能正常完成构造
            return Distributed._lazy_wrapper(survey, cls=cls, dobs=dobs, **kwargs).construct()



from typing import Sequence

import numpy as np
import pandas as pd


def check_components(components):
    if isinstance(components, str):
        return components,
    elif not isinstance(components, Sequence):
        return tuple(components)
    else:
        return components


def format_numpy(data: np.ndarray, components, receiver_locations=None):
    components = check_components(components)
    ret = data.reshape(-1, len(components))
    if receiver_locations is not None:
        ret = np.c_[receiver_locations, ret]

    return ret


def format_pandas(data, components, receiver_locations=None):
    components = check_components(components)
    ret = format_numpy(data, components, receiver_locations)

    headers = components
    if receiver_locations is not None:
        headers = ('x', 'y', 'z', *headers)

    return pd.DataFrame(ret, columns=headers)



# 注意，不能使用from __future__ import annotations
# 会干扰taichi的类型系统，将ti.template等标注识别为字符串形式
# TODO: 关注taichi是否在该方面有相关更新

import taichi as ti

from .taichi import ti_func, ti_pyfunc


@ti_pyfunc
def ti_use(_):
    pass


@ti_func
def ti_index(i, j, stype: ti.template(), struct_size, array_size):
    """对struct array的索引，返回对应的1d index

    Parameters
    ----------
    i
        第i个struct
    j
        struct的第j个元素
    stype
        struct的内存布局类型
    struct_size
        struct的大小
    array_size
        array的大小

    Returns
    -------
    ret
        1d索引
    """

    if ti.static(stype == ti.Layout.AOS):
        return i * struct_size + j  # [x, y, z, x, y, z, ..., x, y, z]
    else:
        return j * array_size + i  # [x, x, ..., x, y, y, ..., y, z, z, ..., z]



from __future__ import annotations

import contextlib

import numpy as np
from SimPEG import maps
from SimPEG.potential_fields.magnetics import Simulation3DIntegral
from discretize.utils import mkvc

from metalpy.mexin import Mixin, mixin
from metalpy.scab.demag.demagnetization import Demagnetization
from metalpy.scab.demag.factored_demagnetization import FactoredDemagnetization
from metalpy.scab.utils.misc import Field


class DemagedMapping(maps.IdentityMap):
    def __init__(self, simulation: Simulation3DIntegral, factor=None, **kwargs):
        assert simulation.model_type == 'scalar', '`Demaged` needs `scalar` model to work.'

        chiMap = simulation.chiMap

        super().__init__(nP=chiMap.shape[0])

        self.simulation = simulation
        self.factor = factor
        self.chiMap = chiMap

        if factor is not None:
            self.demag_solver = FactoredDemagnetization(n=factor)
        else:
            self.demag_solver = Demagnetization(
                mesh=self.simulation.mesh,
                active_ind=self.simulation.ind_active,
                **kwargs
            )

        self.cached_field = None
        self.cached_chi = None

    @property
    def shape(self):
        return self.nP * 3, self.nP

    def _transform(self, m):
        source_field = self.simulation.survey.source_field

        vec_model = self.get_cache(m, source_field)
        if vec_model is None:
            m = self.chiMap * m
            vec_model = self.demag_solver.dpred(m, source_field=source_field)
            self.cache(vec_model, field=source_field)

        return mkvc(vec_model)

    def inverse(self, D):
        raise NotImplemented

    def deriv(self, m, v=None):
        raise NotImplemented

    def cache(self, chi, field):
        self.cached_chi = np.copy(chi)
        self.cached_field = Field(field)

    def get_cache(self, model, field):
        if (
            isinstance(self.demag_solver, Demagnetization)
            and self.test_cached_field(field)
            and self.test_cached_model(model)
        ):
            return np.copy(self.cached_chi)
        else:
            return None

    def test_cached_field(self, field):
        return self.cached_field == Field(field)

    def test_cached_model(self, model):
        cached_model = self.demag_solver.solver.model
        return cached_model is not None and np.all(cached_model == model)


class DemagedMixin(Mixin):
    def __init__(
            self,
            this: Simulation3DIntegral,
            factor=None,
            **kwargs
    ):
        """初始化退磁计算上下文
        """
        super().__init__(this)
        this.chiMap = DemagedMapping(this, factor=factor, **kwargs)

    @mixin.replaces(keep_orig='orig_fn')
    def linear_operator(self, _, orig_fn):
        with self._switch_model_type():
            return orig_fn()

    @contextlib.contextmanager
    def _switch_model_type(self, this: Simulation3DIntegral | None = None):
        """临时将模型类型切换为向量模型
        """
        this.model_type = 'vector'
        try:
            yield
        finally:
            this.model_type = 'scalar'



from __future__ import annotations

import os

from SimPEG.potential_fields.base import BasePFSimulation

from metalpy.utils.type import undefined
from metalpy.scab.modelling.modelled_mesh import ModelledMesh
from .. import SimulationBuilder
from ..simulation import LinearSimulationBuilder


class BasePFSimulationBuilder(LinearSimulationBuilder):
    def __init__(self, sim_cls: BasePFSimulation):
        super().__init__(sim_cls)
        self._model_mesh: ModelledMesh | None = None

    @SimulationBuilder._supplies('mesh', ['ind_active', 'actInd'])
    def active_mesh(self, mesh, ind_active=undefined):
        if not isinstance(mesh, ModelledMesh):
            ind = ind_active if ind_active != undefined else None
            self._model_mesh = ModelledMesh(mesh, ind_active=ind)
        else:
            self._model_mesh = mesh
            if ind_active == undefined:
                ind_active = mesh.active_cells
            mesh = mesh.mesh

        return mesh, ind_active

    @SimulationBuilder._supplies('store_sensitivities', 'sensitivity_path')
    def store_sensitivities(self, target_or_path: bool | str):
        if target_or_path is True:
            return 'ram', undefined
        elif target_or_path is False:
            return 'forward_only', undefined
        else:
            return 'disk', os.fspath(target_or_path)



from .shape3d import Shape3D
from .infinite_shape import InfiniteShape

from .composition import Composition

from .cuboid import Cuboid
from .ellipsoid import Ellipsoid
from .tunnel import Tunnel
from .prism import Prism
from .obj2 import Obj2
from .bar_framework import BarFramework
from .plate_framework import PlateFramework
from .pillar_framework import PillarFramework

from .full_space import FullSpace
from .surface_topo import SurfaceTopo
Terrain = SurfaceTopo



class Worker:
    def __init__(self, name, weight=1, verbose=True):
        self.name = name
        self.weight = weight
        segments = name.split('-')
        # name中最后一个'-'前的内容 或 name，保证name不为空时，结果不为字符串
        self.group = '-'.join(segments[:-1]) or name

        if self.group == name or not segments[-1].isdigit():
            if verbose:
                print("WARNING: worker's name must follow the format '{group name}-{in-group id}', "
                      "and in-group id is assumed to start from zero, "
                      "if not, something may go wrong.")
            self.id = 0
            self.group = name
        else:
            self.id = int(segments[-1])

    def get_name(self):
        return self.name

    def get_weight(self):
        return self.weight

    def get_group(self):
        return self.group

    def get_in_group_id(self):
        return self.id
    
    def __repr__(self):
        return f'{type(self).__name__}(' \
               f'name={self.get_name()}, ' \
               f'weight={self.get_weight()}' \
               f')'



from __future__ import annotations

import json
from typing import cast, Iterable

from metalpy.scab.modelling.shapes import Prism, Cuboid, BarFramework
from metalpy.utils.file import PathLike, openable


def dumps_ptopo(models: Iterable[Prism, Cuboid, BarFramework]) -> list:
    ret = []
    for model in models:
        if isinstance(model, BarFramework):
            model = model.outline

        if isinstance(model, Cuboid):
            pdef = [[
                [model.x0, model.y0],
                [model.x0, model.y1],
                [model.x1, model.y1],
                [model.x1, model.y0],
            ], model.z0, model.z1]
        elif isinstance(model, Prism):
            pdef = [model.pts.tolist(), model.z0, model.z1]
        else:
            raise ValueError(f'Ptopo format supports only Cuboid and Prism shapes. Got {type(model)} instead.')

        if pdef[1] == 0:  # hmin == 0，则隐藏
            pdef = [pdef[0], pdef[2]]
        ret.append(pdef)

    return ret


def dump_ptopo(models: Iterable[Prism, Cuboid], f=None) -> list:
    """将棱柱体或立方体模型组导出为ptopo格式

    ptopo格式用于存储空间中竖直棱柱体的分布信息，格式定义为

    [
        [[p0, ...], hmin, hmax],  # or
        [[p0, ...], hmax],  # where hmin is assumed to be 0
        ...
    ]

    Parameters
    ----------
    models
        棱柱或立方体
    f
        导出的目标文件句柄或路径，为None时不导出

    Returns
    -------
    ret
        ptopo定义下的该棱柱场景list
    """
    ret = dumps_ptopo(models=models)

    if f is not None:
        if openable(f):
            with open(f, 'w') as file:
                json.dump(ret, file)
        else:
            json.dump(ret, f)

    return ret


def load_ptopo(ptopo: PathLike | list) -> list[Prism]:
    """导入ptopo格式定义的棱柱体模型组

    Parameters
    ----------
    ptopo
        符合ptopo格式定义的list或定义文件的路径，传入字符串时假定为路径，会从指定文件读取

    Returns
    -------
    ret : list[Prism]
        一个由棱柱组成的列表，包含ptopo文件中定义的所有棱柱
    """
    if openable(ptopo):
        with open(ptopo, 'r') as f:
            ptopo = json.load(f)

    models = []
    for pdef in ptopo:
        if len(pdef) == 2:
            pts, hmax = pdef
            hmin = 0
        elif len(pdef) == 3:
            pts, hmin, hmax = pdef
        else:
            raise ValueError('Broken prism definition! Expected to be ((pts...) [, hmin], hmax)')
        models.append(Prism(pts, hmin, hmax))

    return models


class PTopoFormat:
    @staticmethod
    def from_ptopo(
        ptopo: PathLike | list,
    ):
        models = load_ptopo(ptopo=ptopo)

        from metalpy.scab.modelling import Scene
        return Scene.of(*models)

    def to_ptopo(
        self,
        path: PathLike | None = None,
    ):
        from metalpy.scab.modelling import Scene
        scene = cast(Scene, self)
        shapes = cast(Iterable, scene.shapes)

        return dump_ptopo(shapes, f=path)



from typing import Literal

from . import is_array_like
from .array_ops import OperatorDispatcher, ArrayType

UStack = OperatorDispatcher('UStack')


def rstack(*arrays, ignore_index=True):
    """同 `np.r_` ，按行增长方向拼接
    """
    return ustack(*arrays, axis=0, ignore_index=ignore_index)


def cstack(*arrays, ignore_index=False):
    """同 `np.c_` ，按列增长方向拼接
    """
    return ustack(*arrays, axis=1, ignore_index=ignore_index)


def ustack(*arrays, axis: Literal[0, 1] = 0, ignore_index=False):
    """拼接向量或矩阵

    Parameters
    ----------
    arrays
        数组列表
    axis
        拼接轴向
    ignore_index
        忽略索引，主要用于pandas

    Returns
    -------
    ret
        返回获取的列，如果输入为切片范围或索引列表，则返回对应矩阵
    """
    assert axis in (0, 1), '`ustack` works only for axis 0 and 1 (as equivalence to np.r_ and np.c_).'
    if len(arrays) == 1:
        if not is_array_like(arrays[0]):
            arrays = tuple(arrays[0])

    return UStack.dispatch(type(arrays[0]))(tuple(arrays), axis=axis, ignore_index=ignore_index)


@UStack.register(ArrayType.numpy)
def _ustack_np(arrays, axis=0, ignore_index=False):
    import numpy as np
    if axis == 0:
        return np.r_.__getitem__(arrays)
    elif axis == 1:
        return np.c_.__getitem__(arrays)


@UStack.register(ArrayType.pandas)
def _ustack_pd(arrays, axis=0, ignore_index=False):
    import pandas as pd
    return pd.concat(arrays, axis=axis, ignore_index=ignore_index)



from typing import TypeVar, Sequence, Generic, Iterator

T = TypeVar('T')


class Batch(Generic[T]):
    __slots__ = ['_items']

    def __init__(self, items: Sequence[T]):
        """用于批量操作对象

        Parameters
        ----------
        items
            需要执行批量操作的对象集合
        """
        self._items = items

    @staticmethod
    def of(items: Sequence[T]) -> "T | Batch[T]":
        """创建一个批操作对象

        Parameters
        ----------
        items
            需要执行批量操作的对象集合

        Returns
        -------
        batch
            批操作对象

        Notes
        -----
        返回值和直接创建一模一样，但是依靠指定返回值类型来欺骗IDE给出提示
        """
        return Batch(items)

    @property
    def batch_items(self):
        return self._items

    def __len__(self):
        return len(self.batch_items)

    def __iter__(self) -> Iterator[T]:
        yield from self.batch_items

    def __call__(self, *args, **kwargs):
        return Batch([t.__call__(*args, **kwargs) for t in self])

    def __getitem__(self, item):
        return Batch([t.__getitem__(item) for t in self])

    def __getattr__(self, item):
        if item not in Batch.__slots__:
            return Batch([getattr(t, item) for t in self])
        else:
            return getattr(self, item)

    def __setattr__(self, item, val):
        if item not in Batch.__slots__:
            for t in self:
                setattr(t, item, val)
        else:
            super().__setattr__(item, val)



from discretize import TensorMesh

from metalpy.utils.dhash import dhash


def dhash_discretize_mesh(mesh: TensorMesh):
    assert isinstance(mesh, TensorMesh), 'hash_discretize_mesh() currently supports only TensorMesh.'
    origin = mesh.origin
    hs = mesh.h

    return dhash(*origin, *hs)



import warnings

from metalpy.utils.object_path import DottedName, get_qualname, get_module_name
from .recoverable_injector import RecoverableInjector
from .utils import wrap_method_with_target
from .replacement import create_replacement


class Extends(RecoverableInjector):
    def __init__(self, nest, name):
        super().__init__()
        self.nest = nest
        self.name = name

    def __call__(self, func):
        if getattr(self.nest, self.name, None) is not None:
            warnings.warn('Trying to extends to a existing target, may lead to unexpected result.')
        wrapper, is_method = wrap_method_with_target(self.nest, func)

        nest_type = self.nest
        if is_method:
            nest_type = type(self.nest)

        wrapper = create_replacement(wrapper, None, self,
                                     qualname=DottedName(get_qualname(nest_type), self.name),
                                     module_name=get_module_name(nest_type))

        cmd = f'self.nest.{self.name} = wrapper'
        exec(cmd)

        return wrapper

    def rollback(self):
        cmd = f'del self.nest.{self.name}'
        exec(cmd)


def extends(target, name):
    return Extends(target, name)



import pyvista as pv

from .texture import Texture


class Diffuse(Texture):
    def __init__(self, texture=None, texture_path=None):
        super().__init__(texture, texture_path)

    def build(self):
        tex = self.texture
        if tex.shape == (3,):
            return {'color': tex}
        else:
            return {'texture': pv.Texture(tex)}



import warnings
from typing import Sequence

from metalpy.utils.string import format_string_list
from .executor import Executor
from .worker import Worker


class SubExecutor(Executor):
    def __init__(self, parent: Executor, workers: Sequence[Worker]):
        """
        """
        super().__init__()
        self.workers = set(workers)
        self.parent = parent

    def do_submit(self, func, *args, workers=None, **kwargs):
        if workers is not None:
            workers = set(workers)
            diff = workers - self.workers
            if len(diff) != 0:
                warnings.warn(
                    'Unexpected workers found. '
                    + format_string_list([f'`{w.name}`' for w in diff], multiline=True)
                )
        else:
            workers = self.workers
        return self.parent.do_submit(func, *args, workers=workers, **kwargs)

    def get_workers(self):
        return self.workers

    def is_local(self):
        return self.parent.is_local()

    def _gather_single(self, future):
        return self.parent._gather_single(future)

    def scatter(self, data):
        return self.parent.scatter(data)

    def get_worker_context(self):
        return self.parent.get_worker_context()

    def receive_events(self, timeout):
        return self.parent.receive_events(timeout)



import numpy as np
from numpy.testing import assert_equal

from metalpy.scab.modelling import Scene
from metalpy.scab.modelling.shapes import Cuboid, Prism, Ellipsoid, Tunnel


def test_build():
    scene = Scene.of(
        Cuboid([1, 1, 1], size=2),
        Prism([[0, 0], [-2, 0], [-2, 1], [-1, 1]], 1, 3),
        Prism([[0, 0], [-2, 0], [-2, 1], [-1, 1]], 1, 3).rotated(90, 0, 0, degrees=True),
        Ellipsoid.spheroid(1, 3, 0).translated(0, -2, 2),
        Tunnel([-3, 2, 2], 0.5, 1, 3),
        models=[1, 2, 3, 4, 5],
    )

    bounds = np.round(scene.bounds)
    ret = [
        scene.build(cell_size=0.2, bounds=bounds),
        scene.build(cell_size=[0.2, 0.2, 0.2], bounds=bounds),
        scene.build(n_cells=13500, bounds=bounds),
        scene.build(n_cells=[30, 30, 15], bounds=bounds)
    ]
    ref_mesh = ret[0]

    for model_mesh in ret[1:]:
        assert_equal(model_mesh.n_cells, ref_mesh.n_cells)
        assert_equal(model_mesh.mesh.cell_centers, ref_mesh.mesh.cell_centers)
        assert_equal(model_mesh.get_active_model(), model_mesh.get_active_model())



def some_func(x):
    return x


class Clazz:
    def __init__(self):
        pass


class DummyClass:
    def __init__(self):
        print('Dummy here')


class ValueHolder:
    def __init__(self, x=1):
        self.x = x



import itertools
import os
import warnings

import cloudpickle
import numpy as np

from metalpy.utils.type import get_or_default, undefined


def dhash(*objs):
    """确定性哈希，通过将Python原生hash中非确定性的部分替换来实现，例如str，并通过扩展方法支持更多对象

    Parameters
    ----------
    objs
        待哈希的对象

    Returns
    -------
    ret : DHash
        DHash实例
    """
    return DHash(*objs)


def register_dhasher(*hashed_types):
    def wrapper(func):
        for t in hashed_types:
            DHash.hashers[t] = func
        return func
    return wrapper


def register_lazy_dhasher(hashed_type: str):
    def wrapper(func):
        DHash.hasher_creators[hashed_type] = func
        return func
    return wrapper


class DHash:
    hashers = {}
    hasher_creators = {}

    def __init__(self, *objs, convert=True):
        """确定性哈希，通过将Python原生hash中非确定性的部分替换来实现，例如str，并通过扩展方法支持更多对象

        Parameters
        ----------
        objs
            待哈希的对象

        Warnings
        --------
            DHash.__init__本身不会对输入的对象进行任何转换，默认均为dhash可接受对象
            请优先使用dhash()函数
        """
        if convert:
            self.objs = tuple(DHash.convert_to_dhashable(obj) for obj in objs)
        else:
            self.objs = objs

        self._result = None

    @property
    def result(self):
        if self._result is None:
            self._result = hash(self.objs)
        return self._result

    def __iter__(self):
        for obj in self.objs:
            if isinstance(obj, DHash):
                for o in obj:
                    yield o
            else:
                yield obj

    def digest(self):
        # TODO: 使用更合理的哈希函数，Python自带的哈希函数会导致结果存在负数
        return abs(self.result)

    def hexdigest(self, digits=32):
        hash_value = self.digest()
        char_table = '0123456789abcdef'
        tablesize = len(char_table)
        hash_string = ''

        while hash_value != 0 and len(hash_string) < digits:
            c = hash_value % tablesize
            hash_string = char_table[c] + hash_string
            hash_value = hash_value // tablesize

        return hash_string

    def __str__(self):
        return self.hexdigest(6)

    def __dhash__(self):
        return self

    def __hash__(self):
        return self.digest()

    def __eq__(self, other):
        return isinstance(other, DHash) and self.result == other.result

    @staticmethod
    def convert_to_dhashable(obj):
        t = type(obj)
        hasher = get_or_default(DHash.hashers, t, None)

        if hasher == undefined:
            hasher = None  # 已确认不存在lazy hasher
        else:
            if hasher is None:
                hasher = DHash._find_lazy_hasher(t)

        if hasher is None:
            def fallback(x): DHash(x, convert=False)
            hasher = getattr(t, '__dhash__', fallback)
            if hasher == fallback:
                warnings.warn(f'Cannot find dhasher for type `{t.__name__}`,'
                              f' falling back to built-in `hash`.')

        return hasher(obj)

    @staticmethod
    def _find_lazy_hasher(t):
        from metalpy.utils.object_path import ObjectPath
        type_name = str(ObjectPath.of(t))
        hasher = get_or_default(DHash.hasher_creators, type_name, None)

        if hasher is not None:
            hasher = DHash.hashers[t] = hasher
        else:
            DHash.hashers[t] = undefined

        return hasher


@register_dhasher(*itertools.chain(*[np.sctypes[k] for k in np.sctypes if k != 'others']))
@register_dhasher(float, int, bool)
def _hash_basic_type(obj):
    return DHash(obj, convert=False)


@register_dhasher(list)
def _hash_list(obj: list):
    return dhash(*obj)


@register_dhasher(tuple)
def _hash_list(obj: tuple):
    return dhash(*obj)


@register_dhasher(str)
def _hash_str(obj: str):
    from functools import reduce
    with np.errstate(over='ignore'):  # 消除上溢警告
        order = np.uint64(5)  # 解决uint64左移运算符的类型错误 https://github.com/numpy/numpy/issues/2524
        ret = reduce(lambda h, c: np.uint64(ord(c)) + ((h << order) + h), obj, np.uint64(5381))
    return ret


@register_dhasher(bytes)
def _hash_bytes(obj: bytes):
    buf = np.frombuffer(obj, dtype=np.uint8)

    from functools import reduce
    with np.errstate(over='ignore'):
        order = np.uint64(5)
        ret = reduce(lambda h, c: c + ((h << order) + h), buf, np.uint64(5381))
    return ret


@register_dhasher(type(None))
def _hash_none(_: None):
    return 0o0225


@register_dhasher(dict)
def _hash_dict(obj: dict):
    return dhash(*sorted(obj.items(), key=lambda x: x[0]))


@register_dhasher(type)
@register_dhasher(type(_hash_dict))  # function
def _hash_serializable(obj):
    return dhash(cloudpickle.dumps(obj))


@register_dhasher(np.ndarray)
def _hash_array(arr: np.ndarray, n_samples=10, sparse=False):
    if not sparse:
        arr = arr.ravel()
        rand = np.random.RandomState(int(arr[len(arr) // 2]) % 2**32)
        n_samples = min(n_samples, len(arr))

        # 加入shape作为参数防止类似np.ones(100)和np.ones(1000)的冲突
        return dhash((arr.shape, *rand.choice(arr, min(len(arr), n_samples), replace=False),))
    else:
        # 稀疏的数组可能会在采样时碰撞概率较大
        # 但是涉及压缩操作可能会导致效率一定程度上降低
        import blosc2
        packed = blosc2.pack_array2(arr)

        return dhash(packed)


@register_lazy_dhasher('pathlib:WindowsPath')
@register_lazy_dhasher('pathlib:PosixPath')
def _hash_path(path):
    return _hash_str(os.fspath(path))


@register_lazy_dhasher('pandas.core.series:Series')
def _hash_series(arr):
    return _hash_array(arr)



from __future__ import annotations

import os.path
from functools import cached_property
from typing import Union, Iterable

import numpy as np
import pyvista as pv

from metalpy.utils.bounds import Bounds
from metalpy.utils.dhash import dhash
from metalpy.utils.file import openable
from metalpy.utils.model import split_models_in_memory, load_model_file, load_grouped_file, \
    extract_model_list_bounds, dhash_model, AxesLabels, reset_model_axes_unsafe
from metalpy.utils.string import parse_axes_labels
from . import Shape3D
from ..utils.mesh import is_inside_bounds


def check_density(length, dx, nx, _default):
    if dx is not None:
        return dx
    elif nx is not None:
        return length / nx
    else:
        return _default


def check_scale(length, target_size, target_scale, _default=1):
    if target_scale is not None:
        return target_scale
    elif target_size is not None:
        return target_size / length
    else:
        return _default


def check_is_none(val, _default):
    if val is not None:
        return val
    else:
        return _default


def check_scalar_or_list(val, size, _default):
    if val is None:
        return [_default] * size
    elif not hasattr(val, '__iter__'):
        return [val] * size
    else:
        return list(val)


class Obj2(Shape3D):
    SUBDIVIDE_IN_MEMORY = 'memory'
    SUBDIVIDE_FILE = 'file'
    SUBDIVIDE_NONE = False

    def __init__(self, model: Union[str, "pv.DataSet", None],
                 xsize=None, ysize=None, zsize=None, size=None,
                 xscale=None, yscale=None, zscale=None, scale=None,
                 surface_thickness=None, surface_range: float | list[float] | None = None,
                 subdivide: str | bool | None = SUBDIVIDE_NONE,
                 reset_axes: AxesLabels | None = None,
                 ignore_surface_check=False,
                 keep_original_model=True,
                 verbose=True):
        """定义一个基于通用格式的模型

        Parameters
        ----------
        model
            模型路径或PyVista模型对象 (当为None时不主动构造，用于内部的拷贝代码，非必要请勿使用)
        xsize, ysize, zsize, size
            模型尺寸，构造时自动将模型缩放到该尺寸下
        xscale, yscale, zscale, scale
            模型缩放比例，构造时自动根据该比例缩放模型
        surface_thickness
            表面厚度（向内），即surface_range=[-surface_range, 0]
        surface_range : array(2) or number
            表面距离范围，用于选定到表面有向距离在该范围内的网格，
            小于0表示模型内，大于0表示模型外，None表示不作判断，
            如果为单个数r，则自动设置为[0, r]或[r, 0]。
        subdivide
            是否模型细分（非通用）。
            SUBDIVIDE_IN_MEMORY 载入后根据点连接性来模型细分。
            SUBDIVIDE_FILE 对模型文件进行组分割（仅限obj文件）再载入。
        reset_axes
            是否重设模型三轴顺序。
            使用一个长度为3的字符或整数数组，指示每个新轴分别对应的原坐标轴。
            例如：交换y和z轴，则指定"xzy"
        ignore_surface_check
            代表使用内置体素化算法时跳过检测模型封闭性
        keep_original_model
            代表是否保留最初导入的未分割的模型，如果为`False`则不会保留，to_polydata()等涉及获取原始模型的操作会通过pv.merge获得。
            适用于不需要再次访问原始模型的情形。
        verbose
            控制是否输出相关信息
        """
        super().__init__()

        self.ignore_surface_check = ignore_surface_check
        if surface_thickness is not None or surface_range is not None:
            if surface_range is not None:
                if isinstance(surface_range, Iterable):
                    self.surface_range = tuple(surface_range)
                else:
                    if surface_range >= 0:
                        self.surface_range = (0, surface_range)  # 0 ~ r
                    else:
                        self.surface_range = (surface_range, 0)  # r ~ 0
            else:
                if surface_thickness is not None:
                    self.surface_range = [-surface_thickness, 0]
                else:
                    self.surface_range = None
        else:
            self.surface_range = None

        if subdivide is True:
            subdivide = Obj2.SUBDIVIDE_IN_MEMORY

        models = None
        did_subdivision = False
        if model is not None:
            owns_model = False
            if openable(model):
                model_path = model
                owns_model = True  # 从文件导入的model的scale操作可以inplace

                if not subdivide:
                    model = load_model_file(model_path, verbose=verbose)
                    models = [model]
                elif subdivide == Obj2.SUBDIVIDE_IN_MEMORY:  # 导入后通过连通性分析分割模型
                    model = load_model_file(model_path, verbose=verbose)
                    models = split_models_in_memory(model, verbose=verbose)
                    did_subdivision = True
                elif subdivide == Obj2.SUBDIVIDE_FILE:  # 从文件层面切割得到的子模型文件导入到models
                    models = load_grouped_file(model_path, verbose=verbose)
                    if models is None:
                        raise NotImplementedError(
                            f"Model file splitter is not available for {os.path.splitext(model_path)[1]} file. \n"
                            f"Which is {model_path} .")
                    did_subdivision = True
                    model = None
                    keep_original_model = False  # 此时不存在原模型，无法保留
            else:
                # then model should be PyVista object
                if not subdivide:
                    models = [model]
                elif subdivide == Obj2.SUBDIVIDE_IN_MEMORY:  # 导入后通过连通性分析分割模型
                    models = split_models_in_memory(model, verbose=verbose)
                    did_subdivision = True

            owns_parts = did_subdivision  # 指示细分模型是否可以inplace操作

            if reset_axes is not None:
                new_axes = parse_axes_labels(reset_axes, max_length=3)

                for i in range(len(models)):
                    models[i] = reset_model_axes_unsafe(models[i], new_axes=new_axes, inplace=owns_parts)
                owns_parts = True

                if did_subdivision and keep_original_model:
                    model = reset_model_axes_unsafe(model, new_axes=new_axes, inplace=owns_model)
                    owns_model = True

            bounds = extract_model_list_bounds(models)
            xl, yl, zl = bounds[1::2] - bounds[0::2]

            size = check_scalar_or_list(size, 3, None)

            size[0] = check_is_none(xsize, size[0])
            size[1] = check_is_none(ysize, size[1])
            size[2] = check_is_none(zsize, size[2])
            xsize, ysize, zsize = size

            scale = check_scalar_or_list(scale, 3, 1)

            scale[0] = check_scale(xl, xsize, xscale, scale[0])
            scale[1] = check_scale(yl, ysize, yscale, scale[1])
            scale[2] = check_scale(zl, zsize, zscale, scale[2])

            if np.any(np.asarray(scale) != 1):
                for i in range(len(models)):
                    models[i] = models[i].scale(scale, inplace=owns_parts)
                owns_parts = True
                if did_subdivision and keep_original_model:
                    model = model.scale(scale, inplace=owns_model)
                    owns_model = True

                bounds = extract_model_list_bounds(models)

            # 保证载入的模型起始点为0
            for i in range(len(models)):
                models[i] = models[i].translate(-bounds[::2], inplace=owns_parts)
            owns_parts = True
            if did_subdivision and keep_original_model:
                model = model.translate(-bounds[::2], inplace=owns_model)
                owns_model = True

        self.models = models

        if keep_original_model:
            if did_subdivision:
                self.model = model
            else:
                if models is not None:
                    self.model = models[0]

    @property
    def surface_only(self):
        return self.surface_range is not None

    @cached_property
    def model(self):
        """获取原始模型，
        如果构造函数指定`keep_original_model`则来自构造函数，
        否则基于细分的models进行merge获取
        """
        return pv.merge(self.models, merge_points=False)

    @property
    def has_original_model(self):
        """判断model属性是否存在缓存值（即不需要进行生成）
        """
        return Obj2.model.attrname in self.__dict__

    def clear_original_model_cache(self):
        """清空model属性的缓存值
        """
        del self.__dict__[Obj2.model.attrname]

    def get_model(self, copy):
        return self.model.copy(deep=copy)

    def place_impl(self, model, mesh_cell_centers, progress):
        indices = is_inside_bounds(mesh_cell_centers, model.bounds)

        mesh_cell_centers = mesh_cell_centers[indices]
        active_indices = indices[indices]

        poly = pv.PolyData(mesh_cell_centers)
        mesh = pv.UnstructuredGrid(poly.cast_to_unstructured_grid())
        surface = model.extract_surface()

        if self.surface_range:
            surface_distance = mesh.compute_implicit_distance(surface)
            smin, smax = self.surface_range

            dist = surface_distance['implicit_distance']

            if smax is not None:
                active_indices &= dist <= smax

            if smin is not None:
                active_indices &= dist >= smin
        else:
            selection = mesh.select_enclosed_points(surface, tolerance=0.0, check_surface=not self.ignore_surface_check)
            active_indices = selection['SelectedPoints'].view(np.bool_)

        indices[indices] = active_indices

        if progress is not None:
            progress.update(1)

        return indices

    def do_place(self, mesh_cell_centers, progress):
        indices = None
        for model in self.models:
            ret = self.place_impl(model, mesh_cell_centers, progress)
            if indices is None:
                indices = ret
            else:
                indices = np.logical_or(indices, ret)

        return indices

    def __dhash__(self):
        n_samples = np.max((2, 10 // len(self.models)))
        return dhash(super().__dhash__(),
                     self.surface_range,
                     *(dhash_model(m, n_samples) for m in self.models),
                     )

    def do_clone(self, deep=True):
        ret = Obj2(
            None,
            surface_range=self.surface_range
        )
        if not deep and self.has_original_model:
            ret.model = self.get_model(copy=False)

        ret.models = [m.copy(deep=deep) for m in self.models]

        return ret

    @property
    def local_bounds(self):
        ret = Bounds(extract_model_list_bounds(self.models))
        if self.surface_range:
            smax = self.surface_range[1]
            if smax is not None:
                ret.expand(increment=smax)

        return ret

    def to_local_polydata(self):
        return self.get_model(copy=True)

    @property
    def volume(self):
        return sum([m.volume for m in self.models])

    @property
    def area(self):
        return sum([m.area for m in self.models])

    @property
    def n_tasks(self):
        return len(self.models)

    @property
    def progress_manually(self):
        return True



from __future__ import annotations

import itertools
import warnings

import numpy as np
import pyvista as pv
from scipy.spatial.transform import Rotation

from metalpy.utils.ear_clip import ear_clip


class TurtlePath:
    def __init__(self, pts=None, fill=False):
        if pts is None:
            pts = []
        self.pts = pts
        self.fill = fill

    def append(self, p):
        self.pts.append(p)

    @property
    def empty(self):
        return len(self.pts) == 0

    @property
    def available(self):
        if self.fill:
            return self.n_points >= 3
        else:
            return self.n_points >= 2

    @property
    def n_points(self):
        return len(self.pts)


class TurtleFillingContext:
    def __init__(self, turtle: PyVistaTurtle):
        self.turtle = turtle

    def __enter__(self):
        self.turtle.begin_fill()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.turtle.end_fill()

    def __bool__(self):
        return self.turtle.path and self.turtle.path.fill


class PyVistaTurtle:
    def __init__(self):
        self.current_position = np.asarray([0, 0, 0])
        self.current_orientation = np.asarray([0, 1, 0])

        self.path: TurtlePath | None = None
        self.shapes: list[TurtlePath] = []

        self.pendown()

    def forward(self, distance):
        self.goto(self.position() + self.heading() * distance)

    fd = forward

    def backward(self, distance):
        self.forward(-distance)

    back = bk = backward

    def left(self, angle, radians=False):
        self.rotate('z', angle, radians=radians)

    lt = left

    def right(self, angle, radians=False):
        self.rotate('z', -angle, radians=radians)

    rt = right

    def goto(self, pos):
        pos = np.asarray(pos)
        if self.isdown():
            if self.path.empty:
                self.path.append(self.position())
            self.path.append(pos)
        self.current_position = pos

    setpos = setposition = goto

    def set_axis(self, axis, val):
        if isinstance(axis, str):
            axis = ord(axis) - ord('x')
        now_pos = self.position()
        now_pos[axis] = val

        self.goto(now_pos)

    def setx(self, x):
        self.set_axis(0, x)

    def sety(self, y):
        self.set_axis(1, y)

    def setz(self, z):
        self.set_axis(2, z)

    def setheading(self, heading):
        heading = np.asarray(heading)
        self.current_orientation = heading / np.linalg.norm(heading)

    seth = setheading

    def home(self):
        self.goto((0, 0, 0))
        self.seth((1, 0, 0))

    def circle(self, radius):
        raise NotImplemented()

    def rotate(self, axes, angle, radians=False):
        angle = np.atleast_1d(angle)
        r = Rotation.from_euler(axes, angle, degrees=not radians)
        self.setheading(r.apply(self.heading())[0])

    def position(self):
        return self.current_position

    pos = position

    def towards(self):
        raise NotImplemented()

    def xcor(self):
        return self.position()[0]

    def ycor(self):
        return self.position()[1]

    def zcor(self):
        return self.position()[2]

    def heading(self):
        return self.current_orientation

    def pendown(self):
        if self.path is None:
            self.path = TurtlePath(fill=False)

    pd = down = pendown

    def penup(self):
        if self.path is not None:
            if not self.path.empty:
                if self.path.n_points < 2:
                    warnings.warn('Only one point in a path. This should not happen. Ignoring it.')
                elif self.path.fill and self.path.n_points < 3:
                    warnings.warn('Insufficient vertices to fill, ignoring it.')
                else:
                    self.shapes.append(self.path)
            self.path = None

    pu = up = penup

    def isdown(self):
        return self.path is not None

    def filling(self):
        """区别于经典turtle API，PyVistaTurtle通过filling函数额外提供Python上下文管理器风格的filling方式。

        Examples
        --------
        >>> turtle = PyVistaTurtle()
        >>> if turtle.filling():  # 判断是否在filling模式，上下文对象隐式转换为bool值
        >>>     print('Filling??')
        >>> with turtle.filling():  # 通过with语法进入filling上下文
        >>>     if turtle.filling():
        >>>         print('Filling!')
        """
        return TurtleFillingContext(self)

    def begin_fill(self):
        self.penup()
        self.pendown()
        self.path.fill = True

    def end_fill(self):
        self.penup()
        self.pendown()

    def to_polydata(self):
        points = []
        lines = []
        cells = []

        additional_shapes = []
        if self.path.available:
            additional_shapes.append([self.path])

        base_point_id = 0
        for s in itertools.chain(self.shapes, *additional_shapes):
            points.append(s.pts)
            if not s.fill:
                pids = np.arange(base_point_id, base_point_id + s.n_points - 1)
                lines.append(np.c_[np.full_like(pids, 2), pids, pids + 1])
            else:
                triangles = ear_clip(s.pts) + base_point_id
                cells.append(np.c_[np.full(*triangles.shape), triangles])

            base_point_id += s.n_points

        kwargs = {}

        points = np.vstack(points)
        if len(lines) > 0:
            lines = np.vstack(lines)
            kwargs['lines'] = lines
        if len(cells) > 0:
            cells = np.vstack(cells)
            kwargs['cells'] = cells

        return pv.PolyData(
            points,
            **kwargs
        )



from metalpy.mexin.injectors import replaces, after, before, extends
from some_module import some_func, Clazz, DummyClass


@before(some_func)
def before_some_func(x):
    print('before some_func')


@after(some_func)
def after_some_func(x):
    print('after some_func')


@before(Clazz.__init__)
def before_clazz_ctor(self):
    print('before Clazz.__init__')


@extends(Clazz, 'extended_method')
def Clazz_extended_method(self, x):
    return x


@replaces(DummyClass)
def DummyClass_hijack():
    return 1



from __future__ import annotations

import copy
import inspect
import warnings
from typing import Callable, Any, Union

from metalpy.utils.type import undefined


class Arg:
    def __init__(self, name):
        self.name = name

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if isinstance(other, Arg):
            return self.name == other.name
        return False

    def __repr__(self):
        return f'{Arg.__name__}("{self.name}")'

    def __str__(self):
        return self.__repr__()


class ArgSpecKey:
    def __init__(self, key):
        self.key = key


IndexType = Union[str, int, Arg, ArgSpecKey]


class ArgSpecs:
    def __init__(self):
        self.posarg_specs: list[Arg] = []
        self.kwarg_specs = set()
        self.n_required_posargs = 0
        self.required_kwargs = set()

        self.contiguous_posargs_count = 0
        self.bound_posargs: dict[int, Any] = {}
        self.bound_kwargs: dict[Arg, Any] = {}
        self.bound_varargs: list[Any] | None = None
        self.bound_varkw: dict[Arg, Any] | None = None

        self.immutable = False

    @staticmethod
    def of(fn: Callable):
        ret = ArgSpecs()

        spec = inspect.getfullargspec(fn)

        args = spec.args
        if isinstance(fn, type):
            args = args[1:]

        kwonlyargs = spec.kwonlyargs
        defaults = spec.defaults or []
        kwonlydefaults = spec.kwonlydefaults or {}

        if spec.varargs is not None:
            ret.accepts_varargs = True

        if spec.varkw is not None:
            ret.accepts_varkw = True

        n_required_posargs = len(args) - len(defaults)
        for i, arg_name in enumerate(args):
            ret.append_arg_spec(arg_name, required=i < n_required_posargs, kwonly=False)

        for arg_name in kwonlyargs:
            ret.append_arg_spec(arg_name, required=arg_name not in kwonlydefaults, kwonly=True)

        return ret

    @staticmethod
    def from_class_mro(cls):
        ret = ArgSpecs.of(cls)

        for parent in inspect.getmro(cls)[1:]:
            ret.merge_kwargs(parent)

        return ret

    def merge_kwargs(self, fn):
        """将可调用对象fn的所有允许使用关键字调用的参数合并到当前ArgSpecs实例的kwargs中

        Parameters
        ----------
        fn
            待分析的fn

        Returns
        -------
        self
            更新后的ArgSpecs实例

        Notes
        -----
            仅在当前参数定义接收kwargs有效，否则会报错
        """
        self._check_immutable()

        if not self.accepts_varkw:
            raise ValueError('Merging kwargs into a function without kwargs is semantically forbidden.')

        spec = inspect.getfullargspec(fn)

        args = spec.args
        if isinstance(fn, type):
            args = args[1:]

        kwonlyargs = spec.kwonlyargs
        defaults = spec.defaults or []
        kwonlydefaults = spec.kwonlydefaults or {}

        n_required_posargs = len(args) - len(defaults)
        for i, arg_name in enumerate(args):
            if self.find_arg_key(arg_name) is None:
                self.append_arg_spec(arg_name, required=i < n_required_posargs, kwonly=True)

        for arg_name in kwonlyargs:
            if self.find_arg_key(arg_name) is None:
                self.append_arg_spec(arg_name, required=arg_name not in kwonlydefaults, kwonly=True)

        return self

    @property
    def accepts_varargs(self):
        return self.bound_varargs is not None

    @accepts_varargs.setter
    def accepts_varargs(self, b: bool):
        if b:
            self.bound_varargs = []
        else:
            if self.bound_varargs is not None and len(self.bound_varargs) > 0:
                warnings.warn('Disabling `varargs` with existing bound values.')
            self.bound_varargs = None

    @property
    def accepts_varkw(self):
        return self.bound_varkw is not None

    @accepts_varkw.setter
    def accepts_varkw(self, b: bool):
        if b:
            self.bound_varkw = {}
        else:
            if self.bound_varkw is not None and len(self.bound_varkw) > 0:
                warnings.warn('Disabling `varkw` with existing bound values.')
            self.bound_varkw = None

    @property
    def n_posargs(self):
        return len(self.posarg_specs)

    @property
    def n_kwargs(self):
        return len(self.kwarg_specs)

    @property
    def n_required_kwargs(self):
        return len(self.required_kwargs)

    @property
    def n_required(self):
        return self.n_required_posargs + self.n_required_kwargs

    @property
    def n_args(self):
        return self.n_posargs + self.n_kwargs

    @property
    def n_varargs(self):
        return 0 if not self.accepts_varargs else len(self.bound_varargs)

    @property
    def n_varkw(self):
        return 0 if not self.accepts_varkw else len(self.bound_varkw)

    @property
    def is_posargs_contiguous(self):
        """指示当前绑定的位置参数是否连续（即每个位置参数之前的所有参数都已给定）
        """
        return self.contiguous_posargs_count == len(self.bound_posargs)

    @property
    def is_satisfied(self):
        return self.is_posargs_satisfied and self.is_kwargs_satisfied

    @property
    def is_posargs_satisfied(self):
        return self.contiguous_posargs_count > self.n_required_posargs

    @property
    def is_all_posargs_provided(self):
        return self.contiguous_posargs_count == self.n_posargs

    @property
    def is_kwargs_satisfied(self):
        return len(self.missing_kwargs) == 0

    @property
    def missing_posargs(self):
        return self.posarg_specs[self.contiguous_posargs_count:self.n_required_posargs]

    @property
    def missing_kwargs(self):
        return self.required_kwargs.difference(set(self.bound_kwargs.keys()))

    def __contains__(self, item):
        ret = self.find_arg_key(item)
        if ret is not None:
            return True
        else:
            return False

    def append_arg_spec(self, name, kwonly=False, required=True):
        self._check_immutable()
        arg = Arg(name)
        if kwonly:
            if required:
                self.required_kwargs.add(arg)
            self.kwarg_specs.add(arg)
        else:
            if required:
                self.n_required_posargs += 1
            self.posarg_specs.append(arg)

    def push_args(self, *args):
        self._mark_immutable()

        if len(args) == 0:
            return

        if not self.is_posargs_contiguous:
            raise RuntimeError('Pushing args to noncontiguous posargs is not allowed.')

        n_seq_args = self.contiguous_posargs_count
        n_new_args = len(args)

        if not self.accepts_varargs and n_seq_args + n_new_args > self.n_posargs:
            raise ValueError(f'Pushing too many values to fixed-length posargs[{self.n_posargs}]'
                             f' (`accepts_varargs` == False).')

        n_non_varargs = min(n_new_args, self.n_posargs - n_seq_args)
        for i, arg in zip(
            range(n_seq_args, n_seq_args + n_non_varargs),
            args[:n_non_varargs]
        ):
            self._bind_arg_by_index_unsafe(i, arg)

        self.contiguous_posargs_count = n_seq_args + n_non_varargs

        if n_non_varargs < n_new_args:
            self._extend_varargs(args[n_non_varargs:])

    def bind_kwargs(self, **kwargs):
        # self._mark_immutable()

        for k, v in kwargs.items():
            self.bind_arg(k, v)

    def bind_arg(self, name_or_index: IndexType, value):
        self._mark_immutable()

        name_or_index = self.find_arg_key(name_or_index, raises=True, allow_vars=True)
        self._bind_or_get_value_unsafe(name_or_index, value)

    def get_bound_arg(self, name_or_index: IndexType, default=undefined, pop=False):
        name_or_index = self.find_arg_key(name_or_index, default=None, raises=default == undefined)
        if name_or_index is None:
            return default
        else:
            ret = self._bind_or_get_value_unsafe(name_or_index)
            if pop:
                self._remove_value_unsafe(name_or_index)
            return ret

    def pop_bound_arg(self, name_or_index: IndexType, default=undefined):
        return self.get_bound_arg(name_or_index=name_or_index, default=default, pop=True)

    def remove(self, name_or_index: IndexType):
        name_or_index = self.find_arg_key(name_or_index, raises=True)
        self._remove_value_unsafe(name_or_index)

    def call(self, func):
        posargs, kwargs = self.build_all_args()
        return func(*posargs, **kwargs)

    def build_all_args(self):
        posargs = []

        missing_posargs = self.missing_posargs
        missing_kwargs = list(self.missing_kwargs)
        missing_tips = []

        if len(missing_posargs) > 0:
            missing_tips.append(f'posarg(s) `{"`, `".join([m.name for m in missing_posargs])}`')

        if len(missing_kwargs) > 0:
            missing_tips.append(f'kwarg(s) `{"`, `".join([m.name for m in missing_kwargs])}`')

        if len(missing_posargs) + len(missing_kwargs) > 0:
            missing_str = f'Missing required {" and ".join(missing_tips)}.'
            raise ValueError(missing_str, missing_posargs, missing_kwargs)

        for i in range(self.contiguous_posargs_count):
            posargs.append(self.bound_posargs[i])

        if self.accepts_varargs and self.n_varargs > 0:
            if not self.is_all_posargs_provided:
                raise ValueError('Cannot bind varargs when posargs are not all specified.',
                                 self.posarg_specs[self.contiguous_posargs_count:self.n_posargs],
                                 [])
            else:
                posargs += self.bound_varargs

        kwargs = {k.name: v for k, v in self.bound_kwargs.items()}

        for i in range(self.contiguous_posargs_count, self.n_posargs):
            if i in self.bound_posargs:
                kwargs[self.posarg_specs[i].name] = self.bound_posargs[i]

        if self.accepts_varkw:
            for k, v in self.bound_varkw.items():
                kwargs[k.name] = v

        return posargs, kwargs

    def find_arg_key(self, name_or_index: IndexType, default=None, raises=False, allow_vars=False):
        if isinstance(name_or_index, Arg):
            name_or_index = name_or_index.name

        if isinstance(name_or_index, ArgSpecKey):
            return name_or_index
        elif isinstance(name_or_index, int):
            if self.n_posargs < name_or_index:
                if not self.accepts_varargs:
                    raise KeyError(f'Index `{name_or_index}` out of range of fixed-length posargs[{self.n_posargs}]'
                                   f' (`accepts_varargs` == False).')
                elif not allow_vars:
                    raise KeyError(f'Index `{name_or_index}` out of range of'
                                   f' bound posargs[{self.contiguous_posargs_count}]')
            return ArgSpecKey(name_or_index)
        elif isinstance(name_or_index, str):
            arg = Arg(name_or_index)
            if arg in self.kwarg_specs:
                return ArgSpecKey(name_or_index)
            else:
                try:
                    index = self.posarg_specs.index(arg)
                    return ArgSpecKey(index)
                except ValueError:
                    pass

            if self.accepts_varkw:
                if arg in self.bound_varkw or allow_vars:
                    return ArgSpecKey(name_or_index)

        if raises:
            raise KeyError(f'`{name_or_index}` does not match any of arg specs.')
        else:
            return default

    def clear(self):
        self._mark_immutable(False)

    def clone_specs(self):
        """只复制参数定义本身，不复制参数的绑定值

        Returns
        -------
        ret
            新的实例
        """
        ret = ArgSpecs()

        ret.posarg_specs = self.posarg_specs
        ret.kwarg_specs = self.kwarg_specs
        ret.n_required_posargs = self.n_required_posargs
        ret.required_kwargs = self.required_kwargs

        ret.accepts_varargs = self.accepts_varargs
        ret.accepts_varkw = self.accepts_varkw

        return ret

    def clone(self):
        """复制参数定义和绑定值

        Returns
        -------
        ret
            新的实例
        """
        ret = self.clone_specs()

        ret.bound_posargs = copy.copy(self.bound_posargs)
        ret.bound_kwargs = copy.copy(self.bound_kwargs)
        ret.bound_varargs = copy.copy(self.bound_varargs)
        ret.bound_varkw = copy.copy(self.bound_varkw)
        ret.contiguous_posargs_count = self.contiguous_posargs_count

        return ret

    def _check_contiguous(self):
        while (self.contiguous_posargs_count > 0 and
               self.contiguous_posargs_count not in self.bound_posargs):
            self.contiguous_posargs_count -= 1
        while self.contiguous_posargs_count in self.bound_posargs:
            self.contiguous_posargs_count += 1

    def _extend_varargs(self, new_varargs):
        self.bound_varargs.extend(new_varargs)

    def _bind_arg_by_index_unsafe(self, index, val=undefined):
        if index < self.n_posargs:
            if val == undefined:
                return self.bound_posargs[index]
            else:
                self.bound_posargs[index] = val
        else:
            if index - self.n_posargs != self.n_varargs:
                warnings.warn('Varargs do not support binding uncontiguously by index. The value will be appended.')

            if val == undefined:
                return self.bound_varargs[index]
            else:
                self.bound_varargs.append(val)

    def _bind_kwarg_by_name_unsafe(self, name, val=undefined):
        name = Arg(name)
        if name in self.kwarg_specs:
            if undefined == val:
                return self.bound_kwargs[name]
            else:
                self.bound_kwargs[name] = val
        elif self.accepts_varkw:
            if undefined == val:
                return self.bound_varkw[name]
            else:
                self.bound_varkw[name] = val

    def _bind_or_get_value_unsafe(self, key, value=undefined):
        if isinstance(key, ArgSpecKey):
            key = key.key

        if isinstance(key, str):
            if undefined == value:
                return self._bind_kwarg_by_name_unsafe(key)
            else:
                self._bind_kwarg_by_name_unsafe(key, value)
        elif isinstance(key, int):
            if undefined == value:
                return self._bind_arg_by_index_unsafe(key)
            else:
                self._bind_arg_by_index_unsafe(key, value)
                self._check_contiguous()

    def _remove_value_unsafe(self, key):
        if isinstance(key, ArgSpecKey):
            key = key.key

        if isinstance(key, str):
            del self.bound_kwargs[Arg(key)]
        elif isinstance(key, int):
            del self.bound_posargs[key]
            self._check_contiguous()

    def _mark_immutable(self, b=True):
        """指示该ArgSpecs实例已经开始绑定值，不再允许修改参数定义
        """
        self.immutable = b

    def _check_immutable(self):
        """检查该ArgSpecs实例是否允许修改参数定义
        """
        if self.immutable:
            raise RuntimeError(f'Modifying arg spec when there are bound values is not allowed.')

    def __repr__(self):
        args = []
        for i in range(self.n_required_posargs):
            args.append(self.posarg_specs[i].name)

        for i in range(self.n_required_posargs, self.n_posargs):
            args.append(f"[{self.posarg_specs[i].name}]")

        if self.accepts_varargs:
            args.append("*args")
        else:
            args.append("*")

        for arg in self.required_kwargs:
            args.append(arg.name)

        for arg in self.kwarg_specs.difference(self.required_kwargs):
            args.append(f'[{arg.name}]')

        if args[-1] == '*':
            # 即没有仅关键字参数
            args.pop()

        if self.accepts_varkw:
            args.append("**kwargs")

        return f'ArgSpecs({", ".join(args)})'

    def get_func_repr(self, func):
        """将参数绑定到函数，获取函数调用的repr表示

        Parameters
        ----------
        func
            函数或函数名

        Returns
        -------
        repr
            给定函数绑定参数后的repr表示
        """
        if not isinstance(func, str):
            func = func.__name__

        posargs, kwargs = self.build_all_args()

        args = []
        for arg in posargs:
            args.append(repr(arg))
        for key, arg in kwargs.items():
            args.append(f'{key}={repr(arg)}')

        return f'{func}({", ".join(args)})'



import inspect
from typing import Union, Callable, Iterable

from metalpy.utils.object_path import ObjectPath
from metalpy.utils.type import get_or_default

type_or_str = Union[type, str]
type_desc = Union[type_or_str, Callable]


class TypeMap:
    def __init__(self, allow_match_parent=False):
        """用于实现类型间的非直接关联

        使用字符串来关联源类型和目标类型，实现在不import源类型和目标类型的情况下进行关联

        支持扩展对象路径语法，可以使用 '/' 分割模块和模块内路径，从而消除歧义
        
        Parameters
        ----------
        allow_match_parent
            是否禁止在找不到直接匹配时递归寻找最近父类的匹配

        See Also
        --------
            get_object_by_path, split_object_path: 扩展对象路径语法

        Notes
        -----
            TypeMap为了规避不必要的import，选择支持使用字符串来表示类型，但这会导致维护困难

            因此一切使用TypeMap的代码应添加相应的测试，请添加测试用例使用assert_all_types_exists验证TypeMap中所定义的类型均有效且存在
        """
        self.registries: dict[ObjectPath, type_desc] = {}
        self.allow_match_parent = allow_match_parent

    def map(self, type_key: type_or_str, type_val: type_desc):
        """注册一个类型间的映射

        Parameters
        ----------
        type_key
            作为key的类型，可以是类型或类型路径
        type_val
            作为val的类型，如果为字符串，则会在需要时动态载入，如果为函数，则会在获取时调用并返回结果
        """
        if isinstance(type_val, str):
            type_val = ObjectPath.of(type_val)
        self.registries[ObjectPath.of(type_key)] = type_val

    def strict_get(self, type_key, default=None) -> type_desc:
        """严格模式下获取key类型对应的原始类型值，不会匹配父类，不会将字符串或函数转换为具体类

        Parameters
        ----------
        type_key
            键类型，如果为类型且无匹配，则会获取对应路径再次尝试匹配
        default
            如果无匹配的返回值

        Returns
        -------
        ret
            获取的类型
        """
        ret = get_or_default(self.registries, ObjectPath.of(type_key), _default=None)
        if ret is None:
            ret = default

        return ret

    def get(self, type_key, default=None) -> type:
        """获取key类型对应的类型，如果strict_mode为False则会匹配父类

        Parameters
        ----------
        type_key
            键类型，如果为类型且无匹配，则会获取对应路径再次尝试匹配
        default
            如果无匹配的返回值

        Returns
        -------
        ret
            获取的类型
        """
        ret = self.strict_get(type_key, default=default)

        if ret is None:
            if self.allow_match_parent and isinstance(type_key, type):
                for parent_type in inspect.getmro(type_key)[1:]:
                    ret = self.strict_get(parent_type, default=default)
            else:
                ret = default

        if isinstance(ret, ObjectPath):
            ret = ret.resolve()
        elif callable(ret):
            ret = ret()

        return ret

    def keys(self) -> Iterable[ObjectPath]:
        """获取TypeMap的所有作为键的类型

        Returns
        -------
        ret
            获取TypeMap的所有原始键
        """
        return self.registries.keys()

    def values(self) -> Iterable[type_desc]:
        """获取TypeMap的所有作为值的类型，不会转换为实际的值

        Returns
        -------
        ret
            获取TypeMap的所有原始值
        """
        return self.registries.values()


def assert_all_types_exists(type_map: TypeMap):
    for key in type_map.keys():
        cls = key.resolve()
        assert isinstance(cls, type), f'Key type "{key}" in TypeMap not found.'
        assert ObjectPath.of(cls) == key, f'ObjectPath to type "{cls.__name__}" is not unique, ' \
                                          f'please use "{ObjectPath.of(cls)}" as the key to this type.'

        value = type_map.get(cls, None)
        assert isinstance(value, type), f'Key "{key}" does not map to a valid type.'



import queue
from concurrent.futures import ThreadPoolExecutor

import psutil

from .pool_executor import PoolExecutor
from .worker import Worker


class ThreadExecutor(PoolExecutor):
    def __init__(self, n_units=None):
        """构造基于ThreadPoolExecutor的执行器

        Parameters
        ----------
        n_units
            工作单位数，一般为物理核心数
        """
        if n_units is None:
            n_units = psutil.cpu_count(logical=False)

        pool = ThreadPoolExecutor(max_workers=n_units)
        workers = [Worker(f'thread-{i}', 1) for i in range(n_units)]

        super().__init__(pool_executor=pool, workers=workers)

    def _get_queue(self):
        return queue.Queue()



import warnings
from typing import Union, Iterable

import numpy as np
from SimPEG.potential_fields.magnetics import UniformBackgroundField
from SimPEG.utils import mat_utils


class Field:
    def __init__(self, intensity, inclination=90, declination=0):
        if isinstance(intensity, UniformBackgroundField):
            self.definition = (intensity.amplitude, intensity.inclination, intensity.declination)
        elif isinstance(intensity, Iterable):
            self.definition = (*intensity,)
        else:
            self.definition = (intensity, inclination, declination)

    @property
    def strength(self):
        warnings.warn("strength is deprecated and will be removed, please use `intensity` instead.", DeprecationWarning)
        return self.definition[0]

    @property
    def intensity(self):
        return self.definition[0]

    @property
    def inclination(self):
        return self.definition[1]

    @property
    def declination(self):
        return self.definition[2]

    @property
    def unit_vector(self):
        return mat_utils.dip_azimuth2cartesian(
            self.inclination,
            self.declination,
        ).squeeze()

    @property
    def vector(self):
        return self.unit_vector * self.intensity

    def __iter__(self):
        for v in self.definition:
            yield v

    def __getitem__(self, index):
        return self.definition[index]

    def __len__(self):
        return 3

    def __eq__(self, rhv):
        return self.definition == Field(rhv).definition

    def with_strength(self, strength):
        warnings.warn("strength is deprecated and will be removed, please use intensity instead.", DeprecationWarning)
        return Field(strength, self.inclination, self.declination)

    def with_intensity(self, intensity):
        return Field(intensity, self.inclination, self.declination)


def define_inducing_field(strength, inclination, declination) -> Field:
    """定义一个外部激发场

    Parameters
    ----------
    strength
        场强
    inclination
        场倾角
    declination
        场偏角

    Returns
    -------
    ret
        定义的物理场
    """
    return Field(strength, inclination, declination)


def define_magnetics_survey(receiver_points: np.ndarray,
                            components: Union[str, Iterable[str]],
                            *other_receivers: Iterable[Union[np.ndarray, Union[str, Iterable[str]]]],
                            inducing_field: Field = None):
    """定义一个磁力学的Survey

    Parameters
    ----------
    receiver_points
        接收点
    components
        接收点所需要计算的分量
    other_receivers
        其他接受点和对应分量，可以有多个，按 rp1, comp1, rp2, comp2, ... 的顺序传入
    inducing_field
        外部激发场，如果为None则使用默认值 (50000nT, 90°, 0°)

    Returns
    -------
    ret
        包含接收点和外部场信息的磁力学Survey
    """
    from SimPEG.potential_fields import magnetics

    receiver_list = magnetics.receivers.Point(receiver_points, components=components)
    receiver_list = [receiver_list]

    assert len(other_receivers) % 2 == 0, '*other_receivers* must be paris of receiver_points, components, ' \
                                          'etc. define_magnetics_survey(rp1, comp1, rp2, comp2, rp3, comp3)'
    for receiver_points, components in zip(other_receivers[::2], other_receivers[1::2]):
        receiver_list.append(magnetics.receivers.Point(receiver_points, components=components))

    source_field = magnetics.sources.SourceField(
        receiver_list=receiver_list, parameters=inducing_field
    )
    survey = magnetics.survey.Survey(source_field)

    return survey



def query_utm_crs_from_bound(bounds):
    from pyproj import CRS
    from pyproj.aoi import AreaOfInterest
    from pyproj.database import query_utm_crs_info

    utm_crs_list = query_utm_crs_info(
        datum_name='WGS 84',
        area_of_interest=AreaOfInterest(
            west_lon_degree=bounds[0],
            east_lon_degree=bounds[1],
            south_lat_degree=bounds[2],
            north_lat_degree=bounds[3],
        ),
    )
    assert len(utm_crs_list) == 1, 'The region should not stretch across UTM zones.'

    utm_crs = CRS.from_epsg(utm_crs_list[0].code)
    return utm_crs


def query_utm_transform_from_bounds(bounds):
    from pyproj import CRS, Transformer

    wgs84_crs = CRS.from_string('WGS 84')
    utm_crs = query_utm_crs_from_bound(bounds)
    to_utm = Transformer.from_crs(wgs84_crs, utm_crs, always_xy=True)

    return to_utm



from metalpy.mepa import Worker
from .distributable import Distributable


class NotDistributable(Distributable):
    def __init__(self):
        super().__init__()

    def should_distribute_to(self, worker: Worker):
        return False



import abc
import warnings
from pathlib import Path

import imageio
import numpy as np

from metalpy.utils.model import DataSetLike


class Texture(abc.ABC):
    def __init__(self, texture=None, texture_path=None):
        if texture is not None:
            if isinstance(texture, Path):
                texture_path = texture
                texture = None
            else:
                texture = np.asarray(texture)
        else:
            assert texture_path is not None, 'Either one of `texture` and `texture_path` must be provided.'

        self.texture = texture
        self.texture_path = texture_path

    def build(self):
        return {}

    def apply(self, actor):
        pass

    @property
    def texture(self):
        if self._texture is None:
            self._texture = imageio.v3.imread(self.texture_path)
        return self._texture

    @texture.setter
    def texture(self, tex):
        self._texture = tex

    @property
    def is_map(self):
        return self.texture.ndim > 1

    def bind(self, dataset: DataSetLike, name=None):
        if name is None:
            if self.texture_path is not None:
                name = Path(self.texture_path).stem

        tex = self.texture
        tex_name, shape_name = self.field_name(name)
        dataset.field_data[tex_name] = tex.ravel()
        dataset.field_data[shape_name] = tex.shape

        if self.is_map:
            # 贴图需要额外激活 t_coords
            if name is not None:
                dataset.point_data.active_t_coords_name = name
            else:
                warnings.warn('Unnamed texture encountered, failed to bind t_coord.')

    @staticmethod
    def extract_all(dataset, name=None):
        textures = TextureSet()
        for texture_type in Texture.get_subclasses():
            tex = texture_type.extract(dataset, name=name)
            if not isinstance(tex, EmptyTexture):
                textures.append(tex)

        return textures

    @classmethod
    def extract(cls, dataset, name=None):
        tex = cls.extract_field(dataset, name=name)
        if tex is None:
            return EmptyTexture()

        return cls(tex)

    @classmethod
    def extract_field(cls, dataset, name=None):
        tex = None
        if name is None:
            for tex_name in dataset.field_data:
                tex_name, shape_name = cls.check_field_name(tex_name)
                if tex_name is None:
                    continue
                tex = dataset.field_data.get(tex_name, None)
                if tex is not None:
                    shape = dataset.field_data[shape_name]
                    tex = tex.reshape(shape)
                    break
        else:
            tex_name, shape_name = cls.field_name(name)
            tex = dataset.field_data[tex_name]
            shape = dataset.field_data[shape_name]
            tex = tex.reshape(shape)

        return tex

    @classmethod
    def field_name(cls, name):
        prefix = cls.__name__
        return f'{prefix}[{name}]', f'{prefix}Shape[{name}]'

    @classmethod
    def check_field_name(cls, tex_name: str):
        prefix = cls.__name__
        if tex_name.startswith(prefix):
            name = tex_name.split('[', maxsplit=1)[1][:-1]
            return f'{prefix}[{name}]', f'{prefix}Shape[{name}]'
        else:
            return None, None

    @classmethod
    def get_subclasses(cls):
        # 请不要整什么奇怪的继承关系，谢谢
        for sub in cls.__subclasses__():
            yield sub
            yield from sub.get_subclasses()


class EmptyTexture(Texture):
    def __init__(self):
        super().__init__(texture=[0])

    def build(self):
        return {}

    @classmethod
    def extract(cls, dataset, name=None):
        return EmptyTexture()


class TextureSet(list[Texture]):
    def build(self):
        ret = {}
        for tex in self:
            ret.update(tex.build())

        return ret

    def apply(self, actor):
        for tex in self:
            tex.apply(actor)

    def bind(self, dataset: DataSetLike, name=None):
        for tex in self:
            tex.bind(dataset, name=name)



from abc import ABC, abstractmethod


class RecoverableInjector(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def __call__(self, func):
        pass

    @abstractmethod
    def rollback(self):
        pass



from __future__ import annotations

import abc
import warnings
from pathlib import Path
from typing import Iterable

import numpy as np
import pyvista as pv

from metalpy.utils.file import openable, PathLike
from metalpy.utils.type import get_or_default
from .textures import TextureSet, Diffuse, Ambient, Specular, Alpha


class TextureHelper:
    def __init__(self):
        self.readers: dict[str, TextureReader] = {}

    def get_reader(self, fmt):
        reader = self.readers.get(fmt, None)
        if reader is None:
            reader = TextureReader.find(fmt)()
            self.readers[fmt] = reader

        return reader

    def bind_texture(self, model, model_path):
        model_path = Path(model_path)
        return self.get_reader(model_path.suffix).bind_texture(model, cwd=model_path.parent)


class TextureReader(abc.ABC):
    _Readers: dict[str, type[TextureReader]] = {}

    @abc.abstractmethod
    def bind_texture(self, model, cwd):
        pass

    @staticmethod
    def of(*formats):
        def wrapper(cls):
            for fmt in formats:
                TextureReader._Readers[fmt] = cls

            return cls

        return wrapper

    @staticmethod
    def find(fmt) -> type[TextureReader]:
        ret = TextureReader._Readers.get(fmt, None)
        if ret is None:
            raise RuntimeError(f'Reading texture for `{fmt}` is currently not supported.')
        else:
            return ret


class DictWithFallback:
    def __init__(self, fallbacks):
        self.dict = {}
        self.fallbacks = fallbacks

    def __setitem__(self, item, value):
        self.dict[item] = value

    def __getitem__(self, item):
        return get_or_default(self.dict, item, supplier=lambda: self.fallbacks[item])

    def get(self, *args, fallback=False, **kwargs):
        if fallback:
            kwargs['supplier'] = lambda: self.fallbacks.get(*args, **kwargs)
        return get_or_default(self.dict, *args, **kwargs)


@TextureReader.of('.obj')
class ObjTextureReader(TextureReader):
    def bind_texture(self, model, cwd):
        return ObjTextureReader.bind_mtl_textures(obj_mesh=model, cwd=cwd)

    @staticmethod
    def bind_mtl_textures(obj_mesh, mtl_path: PathLike | Iterable[PathLike] = None, cwd=None):
        """Modified from @109021017's comment:
        https://github.com/pyvista/pyvista-support/issues/514#issuecomment-1021260384
        """
        if openable(mtl_path):
            mtl_paths = [Path(mtl_path)]
        elif mtl_path is None:
            mtl_paths = [Path(cwd) / p for p in obj_mesh.field_data['MaterialLibraries']]
        else:
            mtl_paths = [Path(p) for p in mtl_path]

        texture_mappings = {}
        for mtl_path in mtl_paths:
            if not mtl_path.exists():
                return obj_mesh

            # parse the mtl file
            for mtl_name, material in ObjTextureReader.parse_mtl(mtl_path):
                textures = TextureSet()
                texture_mappings[mtl_name] = textures

                illum = material['illum']

                if illum >= 0:
                    # Color on
                    kd = material.get('Kd', None)
                    if kd is not None:
                        if isinstance(kd, float):
                            kd = [kd] * 3
                        textures.append(Diffuse(kd))

                if illum >= 1:
                    # Ambient on
                    ka = material.get('Ka', None)
                    if ka is not None:
                        if isinstance(ka, float):
                            ka = [ka] * 3
                        textures.append(Ambient(ka))

                if illum >= 2:
                    # Highlight on
                    ks = material.get('Ks', None)
                    if ks is not None:
                        if isinstance(ks, float):
                            ks = [ks] * 3
                        textures.append(Specular(ks, material['Ns']))

                d = material.get('d', None)
                if d is not None:
                    textures.append(Alpha(d))

                tr = material.get('Tr', None)
                if tr is not None:
                    textures.append(Alpha(1 - tr))

        n_textured = len(texture_mappings)

        if n_textured < 1:
            model = obj_mesh
        elif n_textured == 1:
            model = obj_mesh
            for name, tex in texture_mappings.items():
                if tex is not None:
                    tex.bind(model, name=name)
                    break
        else:
            material_ids = obj_mesh.cell_data['MaterialIds']
            model = pv.MultiBlock()

            materials = obj_mesh.field_data.get('MaterialNames', None)
            if materials is None:
                materials = obj_mesh.cell_data.get('Materials', None)
            if materials is None:
                warnings.warn('No materials list found.'
                              ' Using texture names, which may lead to unexpected result.')
                materials = list(texture_mappings.keys())
                materials.sort()

            for i in np.unique(material_ids):
                name = materials[i]
                mesh_part = obj_mesh.extract_cells(material_ids == i)
                texture_mappings[name].bind(mesh_part, name=name)
                model[name] = mesh_part

        return model

    MaterialBase = {
        'Ka': [0.0, 0.0, 0.0, 1.0],  # ambient
        'Ke': [0.0, 0.0, 0.0, 1.0],  # emissive
        'Kd': [0.5, 0.5, 0.5, 1.0],  # diffuse
        'Ks': [0.0, 0.0, 0.0, 1.0],  # specular
        'Ns': 0.0,  # specular shininess
        'd': 1.0,  # opacity (d / Tr)
        'illum': 2,  # illumination mode
    }

    @staticmethod
    def make_default_material():
        return DictWithFallback(ObjTextureReader.MaterialBase)

    @staticmethod
    def parse_mtl(mtl_path):
        cwd = mtl_path.parent
        with open(mtl_path) as mtl_file:
            material = None
            mtl_name = None
            for line in mtl_file.readlines():
                parts = line.strip().split()
                if len(parts) < 2:
                    continue

                slot = parts[0]
                params = parts[1:]
                if slot.startswith('#'):
                    pass
                elif slot == 'newmtl':
                    if material is not None:
                        yield mtl_name, material
                    material = ObjTextureReader.make_default_material()
                    mtl_name = params[0]
                else:
                    if slot.startswith('map'):
                        path = params[-1]  # 跳过选项
                        material[slot.split('_')[1]] = cwd / path
                    else:
                        fmt = params[0]
                        if fmt in ['spectral', 'xyz']:
                            warnings.warn(f'Format `spectral` and `xyz` are not supported.')
                        else:
                            rgb = [float(c) for c in params]
                            if len(rgb) == 1:
                                rgb = rgb[0]
                            material[slot] = rgb

            if material is not None:
                yield mtl_name, material



import numpy as np

from metalpy.mepa import SingleTaskAllocator


def test_allocator_completeness():
    """测试分配器能否在存在浮点权重时保证按每个权重slice一次后，不会因舍入误差导致数据剩余
    """
    weights = [0.91, 0.4826, 0.11, 0.6739, 0.759, 0.0001]
    alloc = SingleTaskAllocator(np.arange(100), total=sum(weights))
    for w in weights:
        alloc.slice(w)

    assert alloc.finished



from functools import cached_property, reduce

from metalpy.utils.bounds import union, intersects
from metalpy.utils.dhash import dhash
from . import Shape3D
from ..mix_modes import Mixer, MixMode, dhashable_mixer


class Composition(Shape3D):
    # Composition中只考虑正数意义，因此Max和Min分别为并和交
    Union = MixMode.Max
    Intersects = MixMode.Min

    def __init__(self, *shapes: Shape3D, mix_mode: Mixer = Union):
        super().__init__()
        if len(shapes) > 0:
            self.shapes = list(shapes)
        self.mix_mode = mix_mode

    @property
    def mixer(self):
        return MixMode.dispatch(self.mix_mode)

    @cached_property
    def shapes(self) -> list[Shape3D]:
        return []

    @property
    def n_tasks(self):
        return sum((s.n_tasks for s in self.shapes))

    @property
    def progress_manually(self):
        return True

    def do_place(self, mesh_cell_centers, progress):
        mixer = self.mixer
        ret = None
        for shape in self.shapes:
            indices = shape.place(mesh_cell_centers, progress)
            if ret is None:
                ret = indices
            else:
                ret = mixer(ret, indices)

            if progress is not None and not shape.progress_manually:
                progress.update(shape.n_tasks)

        return ret

    def do_compute_implicit_distance(self, mesh_cell_centers, progress):
        if self.mix_mode == Composition.Intersects:
            mixer = MixMode.max
        else:
            mixer = MixMode.min

        ret = None
        for shape in self.shapes:
            distances = shape.compute_implicit_distance(mesh_cell_centers, progress)
            if ret is None:
                ret = distances
            else:
                ret = mixer(ret, distances)

            if progress is not None and not shape.progress_manually:
                progress.update(shape.n_tasks)

        return ret

    def __len__(self):
        return len(self.shapes)

    def __iter__(self):
        yield from self.shapes

    def __getitem__(self, item):
        return self.shapes[item]

    def __dhash__(self):
        return dhash(super().__dhash__(), *self.shapes, dhashable_mixer(self.mix_mode))

    def do_clone(self, deep=True):
        return Composition(
            *[s.clone(deep=deep) for s in self.shapes],
            mix_mode=self.mix_mode
        )

    @property
    def local_bounds(self):
        if self.mix_mode == Composition.Intersects:
            reducer = intersects
        else:
            reducer = union
        return reduce(reducer, (shape.bounds for shape in self))

    def to_local_polydata(self):
        # TODO: 基于模型布尔操作实现常见混合模式下的对象合并处理
        import pyvista as pv
        shapes = [shape.to_polydata() for shape in self]
        shape = pv.MultiBlock(shapes)

        return shape



from .texture import Texture


class Coefficient(Texture):
    def __init__(self, value):
        super().__init__(value, texture_path=None)

    @property
    def name(self):
        return type(self).__name__.lower()

    def build(self):
        return {
            self.name: self.texture.item()
        }



from typing import Union

from .replaces import Replaces
from .utils import wrap_method_with_target


class After(Replaces):
    def __init__(self, target, nest=None, keep_retval: Union[bool, str] = False):
        super().__init__(target, nest=nest, keep_orig='__orig_fn__', force_unbound=True)
        self.keep_retval = keep_retval

    def __call__(self, func):
        func, is_method = wrap_method_with_target(self.nest, func)
        keep_retval = self.keep_retval
        if keep_retval and isinstance(keep_retval, bool):
            keep_retval = 'retval'

        def wrapper(*args, __orig_fn__=None, **kwargs):
            result = __orig_fn__(*args, **kwargs)

            if keep_retval:
                retval = {keep_retval: result}
                ret = func(*args, **kwargs, **retval)
            else:
                ret = func(*args, **kwargs)

            if ret is not None:
                result = ret
            return result

        return super().__call__(wrapper)


def after(target, nest=None, keep_retval=False):
    """将代码注入到目标函数返回后执行

    Parameters
    ----------
    target
        待注入的目标函数
    nest
        手动指定替换对象的所在空间
    keep_retval
        指示是否将原函数返回值作为参数传入

    Notes
    -----
        返回非空的值可以替换原函数的返回值
        keep_retval=True时，原函数返回值将以 **retval** 的名字传入函数，此时请确保替换的函数包含这个名字的参数
        keep_retval为str时，原函数返回值将以该名字传入函数，此时请确保替换的函数包含该指定名字的参数

    See Also
    --------
        before: 将代码注入到函数之前执行
        replaces: 替换目标函数
    """
    return After(target, nest=nest, keep_retval=keep_retval)
